from contextlib import nullcontext as does_not_raise
from functools import partial

import pandas as pd
from pandas.testing import assert_series_equal

from solarforecastarbiter import datamodel
from solarforecastarbiter.reference_forecasts import persistence
from solarforecastarbiter.conftest import default_observation

import pytest


def load_data_base(data, observation, data_start, data_end):
    # slice doesn't care about closed or interval label
    # so here we manually adjust start and end times
    if 'instant' in observation.interval_label:
        pass
    elif observation.interval_label == 'ending':
        data_start += pd.Timedelta('1s')
    elif observation.interval_label == 'beginning':
        data_end -= pd.Timedelta('1s')
    return data[data_start:data_end]


@pytest.fixture
def powerplant_metadata():
    """1:1 AC:DC"""
    modeling_params = datamodel.FixedTiltModelingParameters(
        ac_capacity=200, dc_capacity=200, temperature_coefficient=-0.3,
        dc_loss_factor=3, ac_loss_factor=0,
        surface_tilt=30, surface_azimuth=180)
    metadata = datamodel.SolarPowerPlant(
        name='Albuquerque Baseline', latitude=35.05, longitude=-106.54,
        elevation=1657.0, timezone='America/Denver',
        modeling_parameters=modeling_params)
    return metadata


@pytest.mark.parametrize('interval_label,closed,end', [
    ('beginning', 'left', '20190404 1400'),
    ('ending', 'right', '20190404 1400'),
    ('instant', None, '20190404 1359')
])
def test_persistence_scalar(site_metadata, interval_label, closed, end):
    # interval beginning obs
    observation = default_observation(
        site_metadata, interval_length='5min', interval_label=interval_label)
    tz = 'America/Phoenix'
    data_index = pd.date_range(
        start='20190404', end='20190406', freq='5min', tz=tz)
    data = pd.Series(100., index=data_index)
    data_start = pd.Timestamp('20190404 1200', tz=tz)
    data_end = pd.Timestamp('20190404 1300', tz=tz)
    forecast_start = pd.Timestamp('20190404 1300', tz=tz)
    forecast_end = pd.Timestamp(end, tz=tz)
    interval_length = pd.Timedelta('5min')

    load_data = partial(load_data_base, data)

    fx = persistence.persistence_scalar(
        observation, data_start, data_end, forecast_start, forecast_end,
        interval_length, interval_label, load_data=load_data)

    expected_index = pd.date_range(
        start='20190404 1300', end=end, freq='5min', tz=tz,
        closed=closed)
    expected = pd.Series(100., index=expected_index)
    assert_series_equal(fx, expected)


@pytest.mark.parametrize('obs_interval_label', ('beginning', 'ending',
                                                'instant'))
@pytest.mark.parametrize('interval_label,closed,end', [
    ('beginning', 'left', '20190406 0000'),
    ('ending', 'right', '20190406 0000'),
    ('instant', None, '20190405 2359')
])
def test_persistence_interval(site_metadata, obs_interval_label,
                              interval_label, closed, end):
    # interval beginning obs
    observation = default_observation(
        site_metadata, interval_length='5min',
        interval_label=obs_interval_label)
    tz = 'America/Phoenix'
    data_index = pd.date_range(
        start='20190404', end='20190406', freq='5min', tz=tz)
    # each element of data is equal to the hour value of its label
    data = pd.Series(data_index.hour, index=data_index, dtype=float)
    if obs_interval_label == 'ending':
        # e.g. timestamp 12:00:00 should be equal to 11
        data = data.shift(1).fillna(0)
    data_start = pd.Timestamp('20190404 0000', tz=tz)
    data_end = pd.Timestamp(end, tz=tz) - pd.Timedelta('1d')
    forecast_start = pd.Timestamp('20190405 0000', tz=tz)
    interval_length = pd.Timedelta('60min')

    load_data = partial(load_data_base, data)

    expected_index = pd.date_range(
        start='20190405 0000', end=end, freq='60min', tz=tz, closed=closed)
    expected_vals = list(range(0, 24))
    expected = pd.Series(expected_vals, index=expected_index, dtype=float)

    # handle permutations of parameters that should fail
    if data_end.minute == 59 and obs_interval_label != 'instant':
        expectation = pytest.raises(ValueError)
    elif data_end.minute == 0 and obs_interval_label == 'instant':
        expectation = pytest.raises(ValueError)
    else:
        expectation = does_not_raise()

    with expectation:
        fx = persistence.persistence_interval(
            observation, data_start, data_end, forecast_start,
            interval_length, interval_label, load_data)
        assert_series_equal(fx, expected)


def test_persistence_interval_missing_data(site_metadata):
    # interval beginning obs
    observation = default_observation(
        site_metadata, interval_length='5min',
        interval_label='ending')
    tz = 'America/Phoenix'
    data_index = pd.date_range(
        start='20190404T1200', end='20190406', freq='5min', tz=tz)
    # each element of data is equal to the hour value of its label
    end = '20190406 0000'
    data = pd.Series(data_index.hour, index=data_index, dtype=float)
    data = data.shift(1)
    data_start = pd.Timestamp('20190404 0000', tz=tz)
    data_end = pd.Timestamp(end, tz=tz) - pd.Timedelta('1d')
    forecast_start = pd.Timestamp('20190405 0000', tz=tz)
    interval_length = pd.Timedelta('60min')

    load_data = partial(load_data_base, data)

    expected_index = pd.date_range(
        start='20190405 0000', end=end, freq='60min', tz=tz, closed='right')
    expected_vals = [None] * 12 + list(range(12, 24))
    expected = pd.Series(expected_vals, index=expected_index, dtype=float)
    fx = persistence.persistence_interval(
        observation, data_start, data_end, forecast_start,
        interval_length, 'ending', load_data)
    assert_series_equal(fx, expected)


@pytest.fixture
def uniform_data():
    tz = 'America/Phoenix'
    data_index = pd.date_range(
        start='20190404', end='20190406', freq='5min', tz=tz)
    data = pd.Series(100., index=data_index)
    return data


@pytest.mark.parametrize(
    'interval_label,expected_index,expected_ghi,expected_ac,obsscale', (
        ('beginning',
         ['20190404 1300', '20190404 1330'],
         [96.41150694741889, 91.6991546408236],
         [99.28349914087346, 98.28165269708589],
         1),
        ('ending',
         ['20190404 1330', '20190404 1400'],
         [96.2818141290749, 91.5132934827808],
         [99.25690632023922, 98.2405479197069],
         1),
        # test clipped at 2x clearsky
        ('beginning',
         ['20190404 1300', '20190404 1330'],
         [1926.5828549018618, 1832.4163238767312],
         [395.9216674046528, 391.9265149579709],
         50)
    )
)
def test_persistence_scalar_index(
        powerplant_metadata, uniform_data, interval_label,
        expected_index, expected_ghi, expected_ac, obsscale):
    # ac_capacity is 200 from above
    observation = default_observation(
        powerplant_metadata, interval_length='5min',
        interval_label='beginning')
    observation_ac = default_observation(
        powerplant_metadata, interval_length='5min',
        interval_label='beginning', variable='ac_power')

    data = uniform_data * obsscale
    tz = data.index.tzinfo
    data_start = pd.Timestamp('20190404 1200', tz=tz)
    data_end = pd.Timestamp('20190404 1300', tz=tz)
    forecast_start = pd.Timestamp('20190404 1300', tz=tz)
    forecast_end = pd.Timestamp('20190404 1400', tz=tz)
    interval_length = pd.Timedelta('30min')

    load_data = partial(load_data_base, data)
    fx = persistence.persistence_scalar_index(
        observation, data_start, data_end, forecast_start, forecast_end,
        interval_length, interval_label, load_data)
    expected_index = pd.DatetimeIndex(expected_index, tz=tz)
    expected = pd.Series(expected_ghi, index=expected_index)
    assert_series_equal(fx, expected, check_names=False)

    fx = persistence.persistence_scalar_index(
        observation_ac, data_start, data_end, forecast_start, forecast_end,
        interval_length, interval_label, load_data)
    expected = pd.Series(expected_ac, index=expected_index)
    assert_series_equal(fx, expected, check_names=False)


def test_persistence_scalar_index_instant_obs_fx(
        site_metadata, powerplant_metadata, uniform_data):
    # instantaneous obs and fx
    interval_length = pd.Timedelta('30min')
    interval_label = 'instant'
    observation = default_observation(
        site_metadata, interval_length='5min', interval_label=interval_label)
    observation_ac = default_observation(
        powerplant_metadata, interval_length='5min',
        interval_label=interval_label, variable='ac_power')
    data = uniform_data
    tz = data.index.tzinfo
    load_data = partial(load_data_base, data)
    data_start = pd.Timestamp('20190404 1200', tz=tz)
    data_end = pd.Timestamp('20190404 1259', tz=tz)
    forecast_start = pd.Timestamp('20190404 1300', tz=tz)
    forecast_end = pd.Timestamp('20190404 1359', tz=tz)
    fx = persistence.persistence_scalar_index(
        observation, data_start, data_end, forecast_start, forecast_end,
        interval_length, interval_label, load_data)
    expected_index = pd.DatetimeIndex(
        ['20190404 1300', '20190404 1330'], tz=tz)
    expected_values = [96.59022431746838, 91.99405501672328]
    expected = pd.Series(expected_values, index=expected_index)
    assert_series_equal(fx, expected, check_names=False)

    fx = persistence.persistence_scalar_index(
        observation_ac, data_start, data_end, forecast_start, forecast_end,
        interval_length, interval_label, load_data)
    expected_values = [99.32046515783028, 98.34762206379594]
    expected = pd.Series(expected_values, index=expected_index)
    assert_series_equal(fx, expected, check_names=False)

    # instant obs and fx, but with offset added to starts instead of ends
    data_start = pd.Timestamp('20190404 1201', tz=tz)
    data_end = pd.Timestamp('20190404 1300', tz=tz)
    forecast_start = pd.Timestamp('20190404 1301', tz=tz)
    forecast_end = pd.Timestamp('20190404 1400', tz=tz)
    fx = persistence.persistence_scalar_index(
        observation, data_start, data_end, forecast_start, forecast_end,
        interval_length, interval_label, load_data)
    expected_index = pd.DatetimeIndex(
        ['20190404 1300', '20190404 1330'], tz=tz)
    expected_values = [96.55340033645147, 91.89662922267517]
    expected = pd.Series(expected_values, index=expected_index)
    assert_series_equal(fx, expected, check_names=False)


def test_persistence_scalar_index_invalid_times_instant(site_metadata):
    data = pd.Series(100., index=[0])
    load_data = partial(load_data_base, data)
    tz = 'America/Phoenix'
    interval_label = 'instant'
    observation = default_observation(
        site_metadata, interval_length='5min', interval_label=interval_label)
    # instant obs that cover the whole interval - not allowed!
    data_start = pd.Timestamp('20190404 1200', tz=tz)
    data_end = pd.Timestamp('20190404 1300', tz=tz)
    forecast_start = pd.Timestamp('20190404 1300', tz=tz)
    forecast_end = pd.Timestamp('20190404 1400', tz=tz)
    interval_length = pd.Timedelta('30min')
    with pytest.raises(ValueError):
        persistence.persistence_scalar_index(
            observation, data_start, data_end, forecast_start, forecast_end,
            interval_length, interval_label, load_data)


@pytest.mark.parametrize('interval_label', ['beginning', 'ending'])
@pytest.mark.parametrize('data_start,data_end,forecast_start,forecast_end', (
    ('20190404 1201', '20190404 1300', '20190404 1300', '20190404 1400'),
    ('20190404 1200', '20190404 1259', '20190404 1300', '20190404 1400'),
    ('20190404 1200', '20190404 1300', '20190404 1301', '20190404 1400'),
    ('20190404 1200', '20190404 1300', '20190404 1300', '20190404 1359'),
))
def test_persistence_scalar_index_invalid_times_interval(
        site_metadata, interval_label, data_start, data_end, forecast_start,
        forecast_end):
    data = pd.Series(100., index=[0])
    load_data = partial(load_data_base, data)
    tz = 'America/Phoenix'
    interval_length = pd.Timedelta('30min')

    # base times to mess with
    data_start = pd.Timestamp(data_start, tz=tz)
    data_end = pd.Timestamp(data_end, tz=tz)
    forecast_start = pd.Timestamp(forecast_start, tz=tz)
    forecast_end = pd.Timestamp(forecast_end, tz=tz)

    # interval average obs with invalid starts/ends
    observation = default_observation(
        site_metadata, interval_length='5min', interval_label=interval_label)
    errtext = "with interval_label beginning or ending"
    with pytest.raises(ValueError) as excinfo:
        persistence.persistence_scalar_index(
            observation, data_start, data_end, forecast_start, forecast_end,
            interval_length, interval_label, load_data)
    assert errtext in str(excinfo.value)


def test_persistence_scalar_index_invalid_times_invalid_label(site_metadata):
    data = pd.Series(100., index=[0])
    load_data = partial(load_data_base, data)
    tz = 'America/Phoenix'
    interval_length = pd.Timedelta('30min')

    interval_label = 'invalid'
    observation = default_observation(
        site_metadata, interval_length='5min', interval_label=interval_label)
    data_start = pd.Timestamp('20190404 1200', tz=tz)
    data_end = pd.Timestamp('20190404 1300', tz=tz)
    forecast_start = pd.Timestamp('20190404 1300', tz=tz)
    forecast_end = pd.Timestamp('20190404 1400', tz=tz)
    with pytest.raises(ValueError) as excinfo:
        persistence.persistence_scalar_index(
            observation, data_start, data_end, forecast_start, forecast_end,
            interval_length, interval_label, load_data)
    assert "invalid interval_label" in str(excinfo.value)


def test_persistence_scalar_index_low_solar_elevation(
        site_metadata, powerplant_metadata):

    interval_label = 'beginning'
    observation = default_observation(
        site_metadata, interval_length='5min', interval_label=interval_label)
    observation_ac = default_observation(
        powerplant_metadata, interval_length='5min',
        interval_label=interval_label, variable='ac_power')

    # at ABQ Baseline, solar apparent zenith for these points is
    # 2019-05-13 12:00:00+00:00     91.62
    # 2019-05-13 12:05:00+00:00     90.09
    # 2019-05-13 12:10:00+00:00     89.29
    # 2019-05-13 12:15:00+00:00     88.45
    # 2019-05-13 12:20:00+00:00     87.57
    # 2019-05-13 12:25:00+00:00     86.66

    tz = 'UTC'
    data_start = pd.Timestamp('20190513 1200', tz=tz)
    data_end = pd.Timestamp('20190513 1230', tz=tz)
    index = pd.date_range(start=data_start, end=data_end,
                          freq='5min', closed='left')

    # clear sky 5 min avg (from 1 min avg) GHI is
    # [0., 0.10932908, 1.29732454, 4.67585122, 10.86548521, 19.83487399]
    # create data series that could produce obs / clear of
    # 0/0, 1/0.1, -1/1.3, 5/5, 10/10, 20/20
    # average without limits is (10 - 1 + 1 + 1 + 1) / 5 = 2.4
    # average with element limits of [0, 2] = (2 + 0 + 1 + 1 + 1) / 5 = 1

    data = pd.Series([0, 1, -1, 5, 10, 20.], index=index)
    forecast_start = pd.Timestamp('20190513 1230', tz=tz)
    forecast_end = pd.Timestamp('20190513 1300', tz=tz)
    interval_length = pd.Timedelta('5min')
    load_data = partial(load_data_base, data)

    expected_index = pd.date_range(
        start=forecast_start, end=forecast_end, freq='5min', closed='left')

    # clear sky 5 min avg GHI is
    # [31.2, 44.5, 59.4, 75.4, 92.4, 110.1]
    expected_vals = [31.2, 44.5, 59.4, 75.4, 92.4, 110.1]
    expected = pd.Series(expected_vals, index=expected_index)

    fx = persistence.persistence_scalar_index(
        observation, data_start, data_end, forecast_start, forecast_end,
        interval_length, interval_label, load_data)
    assert_series_equal(fx, expected, check_less_precise=1, check_names=False)

    expected = pd.Series([0.2, 0.7, 1.2, 1.6, 2., 8.9], index=expected_index)
    fx = persistence.persistence_scalar_index(
        observation_ac, data_start, data_end, forecast_start, forecast_end,
        interval_length, interval_label, load_data)
    assert_series_equal(fx, expected, check_less_precise=1, check_names=False)


@pytest.mark.parametrize("interval_label", [
    'beginning', 'ending'
])
@pytest.mark.parametrize("obs_values,axis,constant_values,expected_values", [
    # constant_values = variable values
    # forecasts = percentiles [%]
    ([0, 0, 0, 20, 20, 20], 'x', [10, 20], [50, 100]),

    # constant_values = percentiles [%]
    # forecasts = variable values
    ([0, 0, 0, 4, 4, 4], 'y', [50], [2]),

    # invalid axis
    pytest.param([0, 0, 0, 4, 4, 4], 'percentile', [-1], None,
                 marks=pytest.mark.xfail(raises=ValueError, strict=True)),
])
def test_persistence_probabilistic(site_metadata, interval_label, obs_values,
                                   axis, constant_values, expected_values):

    tz = 'UTC'
    interval_length = '5min'
    observation = default_observation(
        site_metadata,
        interval_length=interval_length,
        interval_label=interval_label
    )

    data_start = pd.Timestamp('20190513 1200', tz=tz)
    data_end = pd.Timestamp('20190513 1230', tz=tz)
    closed = datamodel.CLOSED_MAPPING[interval_label]
    index = pd.date_range(start=data_start, end=data_end, freq='5min',
                          closed=closed)

    data = pd.Series(obs_values, index=index, dtype=float)
    forecast_start = pd.Timestamp('20190513 1230', tz=tz)
    forecast_end = pd.Timestamp('20190513 1300', tz=tz)
    interval_length = pd.Timedelta('5min')
    load_data = partial(load_data_base, data)

    expected_index = pd.date_range(start=forecast_start, end=forecast_end,
                                   freq=interval_length, closed=closed)

    forecasts = persistence.persistence_probabilistic(
        observation, data_start, data_end, forecast_start, forecast_end,
        interval_length, interval_label, load_data, axis, constant_values
    )
    assert isinstance(forecasts, list)
    for i, fx in enumerate(forecasts):
        pd.testing.assert_index_equal(fx.index, expected_index,
                                      check_categorical=False)

        pd.testing.assert_series_equal(
            fx,
            pd.Series(expected_values[i], index=expected_index, dtype=float)
        )


@pytest.mark.parametrize("obs_values,axis,constant_values,expected_values", [
    # constant_values = variable values
    # forecasts = percentiles [%]
    ([0] * 11 + [20] * 11, 'x', [10, 20], [50, 100]),
    ([0] * 11 + [20] * 11, 'x', [10, 20], [50, 100]),

    # constant_values = percentiles [%]
    # forecasts = variable values
    ([0] * 11 + [4] * 11, 'y', [50], [2]),

    # invalid axis
    pytest.param([0] * 11 + [4] * 11, 'percentile', [-1], None,
                 marks=pytest.mark.xfail(raises=ValueError, strict=True)),

    # insufficient observation data
    pytest.param([5.3, 7.3, 1.4] * 4, 'x', [50], None,
                 marks=pytest.mark.xfail(raises=ValueError, strict=True)),
    pytest.param([], 'x', [50], None,
                 marks=pytest.mark.xfail(raises=ValueError, strict=True))
])
def test_persistence_probabilistic_timeofday(site_metadata, obs_values, axis,
                                             constant_values, expected_values):

    tz = 'UTC'
    interval_label = "beginning"
    interval_length = '1h'
    observation = default_observation(
        site_metadata,
        interval_length=interval_length,
        interval_label=interval_label
    )

    # all observations at 9am each day
    data_end = pd.Timestamp('20190513T0900', tz=tz)
    data_start = data_end - pd.Timedelta("{}D".format(len(obs_values)))
    closed = datamodel.CLOSED_MAPPING[interval_label]
    index = pd.date_range(start=data_start, end=data_end, freq='1D',
                          closed=closed)
    data = pd.Series(obs_values, index=index, dtype=float)

    # forecast 9am
    forecast_start = pd.Timestamp('20190514T0900', tz=tz)
    forecast_end = pd.Timestamp('20190514T1000', tz=tz)
    interval_length = pd.Timedelta('1h')

    load_data = partial(load_data_base, data)

    expected_index = pd.date_range(start=forecast_start, end=forecast_end,
                                   freq=interval_length, closed=closed)

    forecasts = persistence.persistence_probabilistic_timeofday(
        observation, data_start, data_end, forecast_start, forecast_end,
        interval_length, interval_label, load_data, axis, constant_values
    )
    assert isinstance(forecasts, list)
    for i, fx in enumerate(forecasts):
        pd.testing.assert_index_equal(fx.index, expected_index,
                                      check_categorical=False)

        pd.testing.assert_series_equal(
            fx,
            pd.Series(expected_values[i], index=expected_index, dtype=float)
        )


@pytest.mark.parametrize("data_end,forecast_start", [
    # no timezone
    (pd.Timestamp("20190513T0900"), pd.Timestamp("20190514T0900")),

    # same timezone
    (
        pd.Timestamp("20190513T0900", tz="UTC"),
        pd.Timestamp("20190514T0900", tz="UTC")
    ),

    # different timezone
    (
        pd.Timestamp("20190513T0200", tz="US/Pacific"),
        pd.Timestamp("20190514T0900", tz="UTC")
    ),

    # obs timezone, but no fx timezone
    (
        pd.Timestamp("20190513T0900", tz="UTC"),
        pd.Timestamp("20190514T0900")
    ),

    # no obs timezone, but fx timezone
    (
        pd.Timestamp("20190513T0900"),
        pd.Timestamp("20190514T0900", tz="UTC")
    ),
])
def test_persistence_probabilistic_timeofday_timezone(site_metadata, data_end,
                                                      forecast_start):

    obs_values = [0] * 11 + [20] * 11
    axis, constant_values, expected_values = 'x', [10, 20], [50, 100]

    interval_label = "beginning"
    interval_length = '1h'
    observation = default_observation(
        site_metadata,
        interval_length=interval_length,
        interval_label=interval_label
    )

    # all observations at 9am each day
    data_start = data_end - pd.Timedelta("{}D".format(len(obs_values)))
    closed = datamodel.CLOSED_MAPPING[interval_label]
    index = pd.date_range(start=data_start, end=data_end, freq='1D',
                          closed=closed)
    data = pd.Series(obs_values, index=index, dtype=float)

    # forecast 9am
    forecast_end = forecast_start + pd.Timedelta("1h")
    interval_length = pd.Timedelta('1h')

    load_data = partial(load_data_base, data)

    expected_index = pd.date_range(start=forecast_start, end=forecast_end,
                                   freq=interval_length, closed=closed)

    # if forecast without timezone, then use obs timezone
    if data.index.tzinfo is not None and forecast_start.tzinfo is None:
        expected_index = expected_index.tz_localize(data.index.tzinfo)

    forecasts = persistence.persistence_probabilistic_timeofday(
        observation, data_start, data_end, forecast_start, forecast_end,
        interval_length, interval_label, load_data, axis, constant_values
    )
    assert isinstance(forecasts, list)
    for i, fx in enumerate(forecasts):
        pd.testing.assert_index_equal(fx.index, expected_index,
                                      check_categorical=False)

        pd.testing.assert_series_equal(
            fx,
            pd.Series(expected_values[i], index=expected_index, dtype=float)
        )


@pytest.mark.parametrize("interval_label", [
    'beginning', 'ending'
])
@pytest.mark.parametrize("obs_values,axis,constant_values,expected_values", [
    # constant_values = variable values
    # forecasts = percentiles [%]
    ([0] * 15 + [20] * 15, 'x', [10, 20], [50, 100]),

    # constant_values = percentiles [%]
    # forecasts = variable values
    ([0] * 15 + [4] * 15, 'y', [50], [2]),

    ([None] * 30, 'y', [50], [None])
])
def test_persistence_probabilistic_resampling(
    site_metadata,
    interval_label,
    obs_values, axis,
    constant_values,
    expected_values
):

    tz = 'UTC'
    interval_length = '1min'
    observation = default_observation(
        site_metadata,
        interval_length=interval_length,
        interval_label=interval_label
    )

    data_start = pd.Timestamp('20190513 1200', tz=tz)
    data_end = pd.Timestamp('20190513 1230', tz=tz)
    closed = datamodel.CLOSED_MAPPING[interval_label]
    index = pd.date_range(start=data_start, end=data_end, freq='1min',
                          closed=closed)

    data = pd.Series(obs_values, index=index, dtype=float)
    forecast_start = pd.Timestamp('20190513 1230', tz=tz)
    forecast_end = pd.Timestamp('20190513 1300', tz=tz)
    interval_length = pd.Timedelta('5min')
    load_data = partial(load_data_base, data)

    expected_index = pd.date_range(start=forecast_start, end=forecast_end,
                                   freq=interval_length, closed=closed)

    forecasts = persistence.persistence_probabilistic(
        observation, data_start, data_end, forecast_start, forecast_end,
        interval_length, interval_label, load_data, axis, constant_values
    )
    assert isinstance(forecasts, list)
    for i, fx in enumerate(forecasts):
        pd.testing.assert_index_equal(fx.index, expected_index,
                                      check_categorical=False)

        pd.testing.assert_series_equal(
            fx,
            pd.Series(expected_values[i], index=expected_index, dtype=float)
        )


@pytest.mark.parametrize("obs_values,axis,constant_values,expected_values", [
    # constant_values = variable values
    # forecasts = percentiles [%]
    ([0] * 22 + [20] * 22, 'x', [10, 20], [50, 100]),

    # constant_values = percentiles [%]
    # forecasts = variable values
    ([0] * 22 + [4] * 22, 'y', [50], [2]),
])
def test_persistence_probabilistic_timeofday_resample(
    site_metadata,
    obs_values,
    axis,
    constant_values,
    expected_values
):

    tz = 'UTC'
    interval_label = "beginning"
    observation = default_observation(
        site_metadata,
        interval_length='30min',
        interval_label=interval_label
    )

    # all observations at 9am each day
    data_end = pd.Timestamp('20190513T0900', tz=tz)
    data_start = data_end - pd.Timedelta("{}D".format(len(obs_values) / 2))
    closed = datamodel.CLOSED_MAPPING[interval_label]
    index = pd.date_range(start=data_start, end=data_end, freq='30min',
                          closed=closed)
    index = index[index.hour == 9]

    data = pd.Series(obs_values, index=index, dtype=float)

    # forecast 9am
    forecast_start = pd.Timestamp('20190514T0900', tz=tz)
    forecast_end = pd.Timestamp('20190514T1000', tz=tz)
    interval_length = pd.Timedelta('1h')

    load_data = partial(load_data_base, data)

    expected_index = pd.date_range(start=forecast_start, end=forecast_end,
                                   freq=interval_length, closed=closed)

    forecasts = persistence.persistence_probabilistic_timeofday(
        observation, data_start, data_end, forecast_start, forecast_end,
        interval_length, interval_label, load_data, axis, constant_values
    )
    assert isinstance(forecasts, list)
    for i, fx in enumerate(forecasts):
        pd.testing.assert_index_equal(fx.index, expected_index,
                                      check_categorical=False)

        pd.testing.assert_series_equal(
            fx,
            pd.Series(expected_values[i], index=expected_index, dtype=float)
        )


@pytest.mark.parametrize("interval_label", [
    'beginning', 'ending'
])
@pytest.mark.parametrize('axis', ['x', 'y'])
def test_persistence_probabilistic_no_data(
        site_metadata, interval_label, axis):

    tz = 'UTC'
    interval_length = '5min'
    observation = default_observation(
        site_metadata,
        interval_length=interval_length,
        interval_label=interval_label
    )

    data_start = pd.Timestamp('20190513 1200', tz=tz)
    data_end = pd.Timestamp('20190513 1230', tz=tz)
    closed = datamodel.CLOSED_MAPPING[interval_label]

    data = pd.Series([], index=pd.DatetimeIndex([], tz=tz), dtype=float)
    forecast_start = pd.Timestamp('20190513 1230', tz=tz)
    forecast_end = pd.Timestamp('20190513 1300', tz=tz)
    interval_length = pd.Timedelta('5min')
    load_data = partial(load_data_base, data)

    expected_index = pd.date_range(start=forecast_start, end=forecast_end,
                                   freq=interval_length, closed=closed)

    forecasts = persistence.persistence_probabilistic(
        observation, data_start, data_end, forecast_start, forecast_end,
        interval_length, interval_label, load_data, axis, [0.0, 25.0, 50.0]
    )
    assert isinstance(forecasts, list)
    for i, fx in enumerate(forecasts):
        pd.testing.assert_index_equal(fx.index, expected_index,
                                      check_categorical=False)

        pd.testing.assert_series_equal(
            fx,
            pd.Series(None, index=expected_index, dtype=float)
        )



import logging


import numpy as np
import pandas as pd
from pvlib.irradiance import get_extra_radiation


from solarforecastarbiter import pvmodel, datamodel
from solarforecastarbiter.io.api import APISession
from solarforecastarbiter.validation import validator, quality_mapping


logger = logging.getLogger(__name__)


def _validate_timestamp(observation, values):
    return validator.check_timestamp_spacing(
        values.index, observation.interval_length, _return_mask=True)


def _validate_stale_interpolated(observation, values):
    window = validator.stale_interpolated_window(observation.interval_length)
    stale_flag = validator.detect_stale_values(values, window=window,
                                               _return_mask=True)
    interpolation_flag = validator.detect_interpolation(values, window=window,
                                                        _return_mask=True)
    return stale_flag, interpolation_flag


def _solpos_night(observation, values):
    solar_position = pvmodel.calculate_solar_position(
        observation.site.latitude, observation.site.longitude,
        observation.site.elevation, values.index)
    night_flag = validator.check_irradiance_day_night(solar_position['zenith'],
                                                      _return_mask=True)
    return solar_position, night_flag


def _solpos_dni_extra(observation, values):
    solar_position, night_flag = _solpos_night(observation, values)
    dni_extra = get_extra_radiation(values.index)
    timestamp_flag = _validate_timestamp(observation, values)
    return solar_position, dni_extra, timestamp_flag, night_flag


def validate_ghi(observation, values):
    """
    Run validation checks on a GHI observation.

    Parameters
    ----------
    observation : solarforecastarbiter.datamodel.Observation
       Observation object that the data is associated with
    values : pandas.Series
       Series of observation values

    Returns
    -------
    timestamp_flag, night_flag, ghi_limit_flag, ghi_clearsky_flag, cloud_free_flag : pandas.Series
        Integer bitmask series from
        :py:func:`.validator.check_timestamp_spacing`,
        :py:func:`.validator.check_irradiance_day_night`,
        :py:func:`.validator.check_ghi_limits_QCRad`,
        :py:func:`.validator.check_ghi_clearsky`,
        :py:func:`.validator.detect_clearsky_ghi` respectively
    """  # NOQA
    solar_position, dni_extra, timestamp_flag, night_flag = _solpos_dni_extra(
        observation, values)
    clearsky = pvmodel.calculate_clearsky(
        observation.site.latitude, observation.site.longitude,
        observation.site.elevation, solar_position['apparent_zenith'])

    ghi_limit_flag = validator.check_ghi_limits_QCRad(
        values, solar_position['zenith'], dni_extra,
        _return_mask=True)
    ghi_clearsky_flag = validator.check_ghi_clearsky(values, clearsky['ghi'],
                                                     _return_mask=True)
    cloud_free_flag = validator.detect_clearsky_ghi(values, clearsky['ghi'],
                                                    _return_mask=True)
    return (timestamp_flag, night_flag, ghi_limit_flag,
            ghi_clearsky_flag, cloud_free_flag)


def validate_dni(observation, values):
    """
    Run validation checks on a DNI observation.

    Parameters
    ----------
    observation : solarforecastarbiter.datamodel.Observation
       Observation object that the data is associated with
    values : pandas.Series
       Series of observation values

    Returns
    -------
    timestamp_flag, night_flag, dni_limit_flag : pandas.Series
        Integer bitmask series from
        :py:func:`.validator.check_timestamp_spacing`,
        :py:func:`.validator.check_irradiance_day_night`,
        :py:func:`.validator.check_dni_limits_QCRad` respectively
    """
    solar_position, dni_extra, timestamp_flag, night_flag = _solpos_dni_extra(
        observation, values)
    dni_limit_flag = validator.check_dni_limits_QCRad(values,
                                                      solar_position['zenith'],
                                                      dni_extra,
                                                      _return_mask=True)
    return timestamp_flag, night_flag, dni_limit_flag


def validate_dhi(observation, values):
    """
    Run validation checks on a DHI observation.

    Parameters
    ----------
    observation : solarforecastarbiter.datamodel.Observation
       Observation object that the data is associated with
    values : pandas.Series
       Series of observation values

    Returns
    -------
    timestamp_flag, night_flag, dhi_limit_flag : pandas.Series
        Integer bitmask series from
        :py:func:`.validator.check_timestamp_spacing`,
        :py:func:`.validator.check_irradiance_day_night`,
        :py:func:`.validator.check_dhi_limits_QCRad` respectively
    """
    solar_position, dni_extra, timestamp_flag, night_flag = _solpos_dni_extra(
        observation, values)
    dhi_limit_flag = validator.check_dhi_limits_QCRad(values,
                                                      solar_position['zenith'],
                                                      dni_extra,
                                                      _return_mask=True)
    return timestamp_flag, night_flag, dhi_limit_flag


def validate_poa_global(observation, values):
    """
    Run validation checks on a POA observation.

    Parameters
    ----------
    observation : solarforecastarbiter.datamodel.Observation
       Observation object that the data is associated with
    values : pandas.Series
       Series of observation values

    Returns
    -------
    timestamp_flag, night_flag, poa_clearsky_flag : pandas.Series
        Integer bitmask series from
        :py:func:`.validator.check_timestamp_spacing`,
        :py:func:`.validator.check_irradiance_day_night`,
        :py:func:`.validator.check_poa_clearsky` respectively
    """
    solar_position, dni_extra, timestamp_flag, night_flag = _solpos_dni_extra(
        observation, values)
    clearsky = pvmodel.calculate_clearsky(
        observation.site.latitude, observation.site.longitude,
        observation.site.elevation, solar_position['apparent_zenith'])
    aoi_func = pvmodel.aoi_func_factory(observation.site.modeling_parameters)
    poa_clearsky = pvmodel.calculate_poa_effective(
        aoi_func=aoi_func, apparent_zenith=solar_position['apparent_zenith'],
        azimuth=solar_position['azimuth'], ghi=clearsky['ghi'],
        dni=clearsky['dni'], dhi=clearsky['dhi'])
    poa_clearsky_flag = validator.check_poa_clearsky(values, poa_clearsky,
                                                     _return_mask=True)
    return timestamp_flag, night_flag, poa_clearsky_flag


def validate_air_temperature(observation, values):
    """
    Run validation checks on an air temperature observation.

    Parameters
    ----------
    observation : solarforecastarbiter.datamodel.Observation
       Observation object that the data is associated with
    values : pandas.Series
       Series of observation values

    Returns
    -------
    timestamp_flag, night_flag, temp_limit_flag : pandas.Series
        Integer bitmask series from
        :py:func:`.validator.check_timestamp_spacing`,
        :py:func:`.validator.check_irradiance_day_night`,
        :py:func:`.validator.check_temperature_limits` respectively
    """
    timestamp_flag, night_flag = validate_defaults(observation, values)
    temp_limit_flag = validator.check_temperature_limits(
        values, _return_mask=True)
    return timestamp_flag, night_flag, temp_limit_flag


def validate_wind_speed(observation, values):
    """
    Run validation checks on a wind speed observation.

    Parameters
    ----------
    observation : solarforecastarbiter.datamodel.Observation
       Observation object that the data is associated with
    values : pandas.Series
       Series of observation values

    Returns
    -------
    timestamp_flag, night_flag, wind_limit_flag : pandas.Series
        Integer bitmask series from
        :py:func:`.validator.check_timestamp_spacing`,
        :py:func:`.validator.check_irradiance_day_night`,
        :py:func:`.validator.check_wind_limits` respectively
    """
    timestamp_flag, night_flag = validate_defaults(observation, values)
    wind_limit_flag = validator.check_wind_limits(values,
                                                  _return_mask=True)
    return timestamp_flag, night_flag, wind_limit_flag


def validate_relative_humidity(observation, values):
    """
    Run validation checks on a relative humidity observation.

    Parameters
    ----------
    observation : solarforecastarbiter.datamodel.Observation
       Observation object that the data is associated with
    values : pandas.Series
       Series of observation values

    Returns
    -------
    timestamp_flag, night_flag, rh_limit_flag : pandas.Series
        Integer bitmask series from
        :py:func:`.validator.check_timestamp_spacing`,
        :py:func:`.validator.check_irradiance_day_night`,
        :py:func:`.validator.check_rh_limits` respectively
    """
    timestamp_flag, night_flag = validate_defaults(observation, values)
    rh_limit_flag = validator.check_rh_limits(values, _return_mask=True)
    return timestamp_flag, night_flag, rh_limit_flag


def validate_ac_power(observation, values):
    """
    Run a number of validation checks on a daily timeseries of AC power.

    Parameters
    ----------
    observation : solarforecastarbiter.datamodel.Observation
       Observation object that the data is associated with
    values : pandas.Series
       Series of observation values

    Returns
    -------
    timestamp_flag, night_flag, limit_flag : pandas.Series
        Integer bitmask series from
        :py:func:`.validator.check_timestamp_spacing`,
        :py:func:`.validator.check_irradiance_day_night`,
        :py:func:`.validator.check_power_limits`
    """
    solar_position, dni_extra, timestamp_flag, night_flag = _solpos_dni_extra(
        observation, values)
    ac_limit_flag = validator.check_ac_power_limits(
        values, solar_position['apparent_zenith'],
        observation.site.modeling_parameters.ac_capacity, _return_mask=True)
    return timestamp_flag, night_flag, ac_limit_flag


def validate_dc_power(observation, values):
    """
    Run a number of validation checks on a daily timeseries of DC power.

    Parameters
    ----------
    observation : solarforecastarbiter.datamodel.Observation
       Observation object that the data is associated with
    values : pandas.Series
       Series of observation values

    Returns
    -------
    timestamp_flag, night_flag, limit_flag : pandas.Series
        Integer bitmask series from
        :py:func:`.validator.check_timestamp_spacing`,
        :py:func:`.validator.check_irradiance_day_night`,
        :py:func:`.validator.check_power_limits`
    """
    solar_position, dni_extra, timestamp_flag, night_flag = _solpos_dni_extra(
        observation, values)
    dc_limit_flag = validator.check_dc_power_limits(
        values, solar_position['apparent_zenith'],
        observation.site.modeling_parameters.dc_capacity, _return_mask=True)
    return timestamp_flag, night_flag, dc_limit_flag


def validate_defaults(observation, values):
    """
    Run default validation checks on an observation.

    Parameters
    ----------
    observation : solarforecastarbiter.datamodel.Observation
       Observation object that the data is associated with
    values : pandas.Series
       Series of observation values

    Returns
    -------
    timestamp_flag, night_flag : pandas.Series
        Integer bitmask series from
        :py:func:`.validator.check_timestamp_spacing`,
        :py:func:`.validator.check_irradiance_day_night` respectively
    """
    timestamp_flag = _validate_timestamp(observation, values)
    _, night_flag = _solpos_night(observation, values)
    return timestamp_flag, night_flag


def validate_daily_ghi(observation, values):
    """
    Run validation on a daily timeseries of GHI. First,
    all checks of `validate_ghi` are run in addition to
    detecting stale values and interpolation

    Parameters
    ----------
    observation : solarforecastarbiter.datamodel.Observation
       Observation object that the data is associated with
    values : pandas.Series
       Series of observation values

    Returns
    -------
    *ghi_flags, stale_flag, interpolation_flag : pandas.Series
        Integer bitmask series from
        :py:func:`.validate_ghi`,
        :py:func:`.validator.detect_stale_values`,
        :py:func:`.validator.detect_interpolation`
    """
    ghi_flags = validate_ghi(observation, values)
    stale_flag, interpolation_flag = _validate_stale_interpolated(observation,
                                                                  values)
    return (*ghi_flags, stale_flag, interpolation_flag)


def validate_daily_dc_power(observation, values):
    """
    Run validation on a daily timeseries of DC power.

    Parameters
    ----------
    observation : solarforecastarbiter.datamodel.Observation
       Observation object that the data is associated with
    values : pandas.Series
       Series of observation values

    Returns
    -------
    timestamp_flag, night_flag, limit_flag, stale_flag, interpolation_flag : pandas.Series
        Integer bitmask series from
        :py:func:`.validator.check_timestamp_spacing`,
        :py:func:`.validator.check_irradiance_day_night`,
        :py:func:`.validator.check_power_limits`,
        :py:func:`.validator.detect_stale_values`,
        :py:func:`.validator.detect_interpolation`
    """  # NOQA: E501
    timestamp_flag, night_flag, dc_limit_flag = validate_dc_power(observation,
                                                                  values)
    stale_flag, interpolation_flag = _validate_stale_interpolated(observation,
                                                                  values)
    return (timestamp_flag, night_flag, dc_limit_flag, stale_flag,
            interpolation_flag)


def validate_daily_ac_power(observation, values):
    """
    Run a number of validation checks on a daily timeseries of AC power.

    Parameters
    ----------
    observation : solarforecastarbiter.datamodel.Observation
       Observation object that the data is associated with
    values : pandas.Series
       Series of observation values

    Returns
    -------
    timestamp_flag, night_flag, limit_flag, stale_flag, interpolation_flag, clipping_flag : pandas.Series
        Integer bitmask series from
        :py:func:`.validator.check_timestamp_spacing`,
        :py:func:`.validator.check_irradiance_day_night`,
        :py:func:`.validator.check_power_limits`,
        :py:func:`.validator.detect_stale_values`,
        :py:func:`.validator.detect_interpolation`,
        :py:func:`.validator.detect_clipping`
    """  # NOQA: E501
    timestamp_flag, night_flag, ac_limit_flag = validate_ac_power(observation,
                                                                  values)
    stale_flag, interpolation_flag = _validate_stale_interpolated(observation,
                                                                  values)
    clipping_flag = validator.detect_clipping(values, _return_mask=True)
    return (timestamp_flag, night_flag, ac_limit_flag, stale_flag,
            interpolation_flag, clipping_flag)


def validate_daily_defaults(observation, values):
    """
    Run default daily validation checks on an observation.
    Applies the validation for the observation's variable and then
    the stale and interpolated validation. :py:func:`validate_defaults`
    is used if the Observation variable does not have a defined validation
    function.

    Parameters
    ----------
    observation : solarforecastarbiter.datamodel.Observation
       Observation object that the data is associated with
    values : pandas.Series
       Series of observation values

    Returns
    -------
    *variable_immediate_flags, stale_flag, interpolation_flag : pandas.Series
        Integer bitmask series from
        :py:func:`.tasks.validate_{variable}`,
        :py:func:`.validator.detect_stale_values`,
        :py:func:`.validator.detect_interpolation`
    """
    immediate_func = IMMEDIATE_VALIDATION_FUNCS.get(
        observation.variable, validate_defaults)
    immediate_flags = immediate_func(observation, values)
    stale_flag, interpolation_flag = _validate_stale_interpolated(observation,
                                                                  values)
    return (*immediate_flags, stale_flag, interpolation_flag)


IMMEDIATE_VALIDATION_FUNCS = {
    'air_temperature': validate_air_temperature,
    'wind_speed': validate_wind_speed,
    'ghi': validate_ghi,
    'dni': validate_dni,
    'dhi': validate_dhi,
    'poa_global': validate_poa_global,
    'relative_humidity': validate_relative_humidity,
    'ac_power': validate_ac_power,
    'dc_power': validate_dc_power
}
DAILY_VALIDATION_FUNCS = {
    'ghi': validate_daily_ghi,
    'dc_power': validate_daily_dc_power,
    'ac_power': validate_daily_ac_power,
    # no stale/interpolated
    'event': validate_defaults,
    'availability': validate_defaults,
    'curtailment': validate_defaults,
}


def apply_immediate_validation(observation, observation_values):
    """
    Applies the appropriate validation functions to the observation_values.
    Only the USER_FLAGGED flag is propagated if the series has been
    previously validated.

    Parameters
    ----------
    observation : solarforecastarbiter.datamodel.Observation
    observation_values : pandas.DataFrame
        Must have 'value' and 'quality_flag' columns

    Returns
    -------
    pandas.DataFrame
        With the same index as the input and 'quality_flag' updated
        appropriately
    """
    value_series = observation_values['value'].astype(float)
    quality_flags = observation_values['quality_flag'].copy() & 1

    validation_func = IMMEDIATE_VALIDATION_FUNCS.get(
        observation.variable, validate_defaults)
    validation_flags = validation_func(observation, value_series)

    for flag in validation_flags:
        quality_flags |= flag
    quality_flags |= quality_mapping.LATEST_VERSION_FLAG

    quality_flags.name = 'quality_flag'
    observation_values.update(quality_flags)
    return observation_values


def apply_daily_validation(observation, observation_values):
    """Applies the appropriate daily validation functions to the
    observation_values.  Only the USER_FLAGGED flag is propagated if
    the series has been previously validated.

    Parameters
    ----------
    observation : solarforecastarbiter.datamodel.Observation
    observation_values : pandas.DataFrame
        Must have 'value' and 'quality_flag' columns

    Returns
    -------
    pandas.DataFrame
        With the same index as the input and 'quality_flag' updated
        appropriately

    Raises
    ------
    IndexError
        If there are not enough valid points to perform daily validation

    """
    validated = observation_values.sort_index()
    value_series = validated['value'].astype(float)
    if len(value_series.dropna()) < 10:
        raise IndexError(
            'Data series does not have at least 10 datapoints to validate')
    quality_flags = validated['quality_flag'].copy() & 1

    # if the variable has a daily check, run that, else run the
    # immediate validation, else validate timestamps
    validation_func = DAILY_VALIDATION_FUNCS.get(
        observation.variable, validate_daily_defaults)
    validation_flags = validation_func(observation, value_series)

    for flag in validation_flags:
        quality_flags |= flag
    quality_flags |= quality_mapping.DAILY_VALIDATION_FLAG
    quality_flags |= quality_mapping.LATEST_VERSION_FLAG

    quality_flags.name = 'quality_flag'
    validated.update(quality_flags)
    return validated


def apply_validation(observation, observation_values):
    """
    Applies the appropriate daily or immediate validation functions to the
    observation_values depending on the length of the data. If an Aggregate
    object is passed, a warning is logged and the observation_values are
    returned.

    Parameters
    ----------
    observation : solarforecastarbiter.datamodel.Observation
    observation_values : pandas.DataFrame
        Must have 'value' and 'quality_flag' columns

    Returns
    -------
    pandas.DataFrame
        With the same index as the input and 'quality_flag' updated
        appropriately

    Raises
    ------
    TypeError
        If the supplied observations_values is not a DataFrame with a
        DatetimeIndex
    """
    if isinstance(observation, datamodel.Aggregate):
        logger.warning('Cannot apply validation to an Aggregate')
        return observation_values
    data = observation_values.sort_index()
    if (
            not isinstance(data, pd.DataFrame) or
            not isinstance(data.index, pd.DatetimeIndex)
    ):
        raise TypeError('Expect observation_values to have a DatetimeIndex')
    if data.empty:
        return data
    if (
            (data.index[-1] - data.index[0]) >= pd.Timedelta('1d') and
            (len(data['value'].dropna()) > 10)
    ):
        return apply_daily_validation(observation, data)
    else:
        return apply_immediate_validation(observation, data)


def _group_continuous_week_post(session, observation, observation_values):
    # observation_values expected to be sorted
    # observation values already have uneven frequency checked
    gid = quality_mapping.check_if_series_flagged(
        observation_values['quality_flag'], 'UNEVEN FREQUENCY').cumsum()
    # make series of week + year integers to further
    # split data to post at most one week at a time
    # ~10,000 pts of 1min data
    week_int = (gid.index.week + gid.index.year).values
    # combine the continuous groups with groups of weeks
    # gid is unique for each group since week_int and cumsum
    # increase monotonically and are positive
    gid += week_int
    observation_values['gid'] = gid
    for _, group in observation_values.groupby('gid'):
        session.post_observation_values(observation.observation_id,
                                        group[['value', 'quality_flag']],
                                        params='donotvalidate')


def _validate_post(session, observation, start, end):
    logger.info('Validating data for %s from %s to %s',
                observation.name, start, end)
    observation_values = session.get_observation_values(
        observation.observation_id, start, end)
    validated = apply_validation(observation, observation_values)
    return _group_continuous_week_post(
        session, observation, validated)


def _find_unvalidated_time_ranges(session, observation, min_start, max_end):
    """Find the time ranges where the observation data needs to have
    daily validation applied. Extend to next day midnight so daily
    validation can be applied even since it requires >= 1 day of data
    """
    tz = observation.site.timezone
    dates = session.get_observation_values_not_flagged(
        observation_id=observation.observation_id,
        start=min_start,
        end=max_end,
        flag=(
            quality_mapping.DAILY_VALIDATION_FLAG |
            quality_mapping.LATEST_VERSION_FLAG
        ),
        timezone=tz)
    if len(dates) == 0:
        return
    sorted_dates = np.array(sorted(dates))

    def first_last(prev, ind):
        first = pd.Timestamp(dates[prev]).tz_localize(tz)
        last = (pd.Timestamp(dates[ind])
                .tz_localize(tz) + pd.Timedelta('1D'))
        return first, last

    prev = 0
    # find the difference between each date, as integer days
    # subtract one to then use nonzero to find those
    # dates that are not continuous
    breaks = np.diff(sorted_dates).astype('timedelta64[D]') - 1
    discontinuities = np.nonzero(breaks)[0]

    for ind in discontinuities:
        first, last = first_last(prev, ind)
        yield first, last
        prev = ind + 1

    first, last = first_last(prev, -1)
    yield first, last


def _split_validation(session, observation, start, end, only_missing):
    if not only_missing:
        return _validate_post(session, observation, start, end)

    for _start, _end in _find_unvalidated_time_ranges(
            session, observation, start, end):
        _validate_post(session, observation, _start, _end)


def fetch_and_validate_observation(access_token, observation_id, start, end,
                                   only_missing=False, base_url=None):
    """Task that will run immediately after Observation values are
    uploaded to the API to validate the data. If over a day of data is
    present, daily validation will be applied.

    For the last day of a multiday series that only has a partial day's
    worth of data, if `only_missing` is False, the data is evaluated as
    one series and daily validation is applied. If `only_missing` is True,
    any discontinuous periods of data with less than one day of data will
    only have immediate validation applied. If the period is longer than
    a day, the full daily validation is applied.

    Parameters
    ----------
    access_token : str
        Token to access the API
    observation_id : str
        ID of the observation to fetch values and validate
    start : datetime-like
        Start time to limit observation fetch
    end : datetime-like
        End time to limit observation fetch
    only_missing : boolean, default False
        If True, only periods that have not had daily validation applied
        are fetched and validated. Otherwise all data between start and end
        is validated.
    base_url : str, default None
        URL for the API to fetch and post data
    """
    session = APISession(access_token, base_url=base_url)
    observation = session.get_observation(observation_id)
    _split_validation(session, observation, start, end, only_missing)


def fetch_and_validate_all_observations(access_token, start, end,
                                        only_missing=True, base_url=None):
    """
    Run the observation validation for all observations that the user
    has access to in their organization. See further discussion in
    :py:func:`solarforecastarbiter.validation.tasks.fetch_and_validate_all_observations`

    Parameters
    ----------
    access_token : str
        Token to access the API
    start : datetime-like
        Start time to limit observation fetch
    end : datetime-like
        End time to limit observation fetch
    only_missing : boolean, default True
        If True, only periods that have not had daily validation applied
        are fetched and validated. Otherwise all data between start and end
        is validated.
    base_url : str, default None
        URL for the API to fetch and post data

    """
    session = APISession(access_token, base_url=base_url)
    user_info = session.get_user_info()
    observations = [obs for obs in session.list_observations()
                    if obs.provider == user_info['organization']]
    for observation in observations:
        _split_validation(session, observation, start, end, only_missing)



import inspect
import os


from netCDF4 import Dataset
import pandas as pd
import pytest
from requests.exceptions import ChunkedEncodingError


from solarforecastarbiter.io.fetch import arm


TEST_DATA_DIR = os.path.dirname(
    os.path.abspath(inspect.getfile(inspect.currentframe())))
MET_FILE = os.path.join(TEST_DATA_DIR, 'data',
                        'sgpmetE13.b1.20190122.000000.cdf')
IRRAD_FILE = os.path.join(TEST_DATA_DIR, 'data',
                          'sgpqcrad1longC1.c1.20190122.000000.cdf')


test_datastreams = ['ds_1', 'ds_2']
start_date = pd.Timestamp('2019-01-22')
end_date = pd.Timestamp('2019-01-23')


def filenames(*args):
    if args[2] == 'ds_1':
        return ['irrad']
    if args[2] == 'ds_2':
        return ['weather']
    return []


def request_file(*args):
    if args[2] == 'irrad':
        return Dataset(IRRAD_FILE)
    if args[2] == 'weather':
        return Dataset(MET_FILE)


def test_format_date():
    date = pd.Timestamp('2019-01-23T01:01:01Z')
    assert arm.format_date(date) == '2019-01-23'


def mocked_request_get_files(*args, **kwargs):
    class Object:
        pass
    response = Object()
    response.text = '{"files": ["filename1", "filename2"]}'
    return response


@pytest.fixture
def api_key():
    return 'bogus_key'


@pytest.fixture
def user_id():
    return 'user_id'


@pytest.mark.parametrize('stream,variables,start,end', [
    (test_datastreams[0], ['down_short_hemisp',
     'not_real'], start_date, end_date),
    (test_datastreams[1], ['temp_mean'], start_date,
     end_date),
])
def test_fetch_arm(user_id, api_key, stream, variables, start, end, mocker):
    mocker.patch('solarforecastarbiter.io.fetch.arm.list_arm_filenames',
                 side_effect=filenames)
    mocker.patch('solarforecastarbiter.io.fetch.arm.retrieve_arm_dataset',
                 side_effect=request_file)
    data = arm.fetch_arm(user_id, api_key, stream, variables, start, end)
    assert variables[0] in data.columns


@pytest.mark.parametrize('stream,start,end', [
    ('datastream', start_date, end_date)
])
def test_request_file_lists(user_id, api_key, stream, start, end, mocker):
    mocked_get = mocker.patch('solarforecastarbiter.io.fetch.arm.requests.get',
                              side_effect=mocked_request_get_files)
    arm.list_arm_filenames(user_id, api_key, stream, start, end)
    mocked_get.assert_called_with(
        'https://adc.arm.gov/armlive/data/query',
        params={
            'user': f'{user_id}:{api_key}',
            'ds': 'datastream',
            'start': '2019-01-22',
            'end': '2019-01-23',
            'wt': 'json'
        })


def test_request_arm_file(user_id, api_key, mocker):
    mocked_get = mocker.patch('solarforecastarbiter.io.fetch.arm.requests.get')
    arm.request_arm_file(user_id, api_key, 'sgpqcrad1longC1.c1.cdf')
    mocked_get.assert_called_with(
        arm.ARM_FILES_DOWNLOAD_URL,
        params={
            'user': f'{user_id}:{api_key}',
            'file': 'sgpqcrad1longC1.c1.cdf',
        })


def test_extract_arm_variables_exist(mocker):
    nc_file = request_file(None, None, 'irrad')
    extracted = arm.extract_arm_variables(nc_file,
                                          ['down_short_hemisp', 'nonexistent'])
    assert 'down_short_hemisp' in extracted.columns
    assert 'non-existent' not in extracted.columns


def test_extracted_arm_variables_empty(mocker):
    nc_file = request_file(None, None, 'irrad')
    extracted = arm.extract_arm_variables(nc_file,
                                          ['no', 'nein', 'ie', 'non'])
    assert extracted.empty


def test_no_files(user_id, api_key, mocker):
    mocker.patch('solarforecastarbiter.io.fetch.arm.list_arm_filenames',
                 side_effect=filenames)
    mocker.patch('solarforecastarbiter.io.fetch.arm.retrieve_arm_dataset',
                 side_effect=request_file)
    start = end = pd.Timestamp.now()+pd.Timedelta('1 days')
    arm_df = arm.fetch_arm(user_id, api_key, 'ds_no_files',
                           ['down_short_hemisp'], start, end)
    assert arm_df.empty


@pytest.mark.parametrize('num_failures', range(1, 5))
def test_request_arm_file_retries(mocker, num_failures):
    mocked_get = mocker.patch('solarforecastarbiter.io.fetch.arm.requests.get')
    return_values = (d for d in [0, num_failures])

    def get_response(*args, **kwargs):
        call_no = next(return_values)
        if call_no < num_failures:
            raise ChunkedEncodingError
        else:
            response = mocker.MagicMock()
            response.content = 'success'
            return response

    mocked_get.side_effect = get_response
    the_response = arm.request_arm_file('user', 'ley', 'filename')
    assert the_response == 'success'


def test_request_arm_file_failure_after_retries(mocker):
    mocked_get = mocker.patch('solarforecastarbiter.io.fetch.arm.requests.get')
    mocked_get.side_effect = ChunkedEncodingError
    with pytest.raises(ChunkedEncodingError):
        arm.request_arm_file('user', 'ley', 'filename')


def test_fetch_arm_request_file_failure(mocker):
    mocker.patch('solarforecastarbiter.io.fetch.arm.list_arm_filenames',
                 return_value=['afilename'])
    mocker.patch('solarforecastarbiter.io.fetch.arm.retrieve_arm_dataset',
                 side_effect=ChunkedEncodingError)
    mocked_log = mocker.patch('solarforecastarbiter.io.fetch.arm.logger')
    data = arm.fetch_arm('user', 'key', 'stream', ['ghi'], 'start', 'end')
    mocked_log.error.assert_called_with(
            f'Request failed for DOE ARM file afilename')
    assert data.empty



# models_persistence.py?
"""
Functions for persistence forecasts.

Two kinds of persistence are supported:

  1. Persistence of observed values in :py:func:`persistence_scalar` and
     :py:func:`persistence_interval`
  2. Persistence of irradiance or power accounting for solar position in
     :py:func:`persistence_scalar_index` and
     :py:func:`persistence_interval_index` (?).

Users of intraday persistence forecasts will typically want to use
:py:func:`persistence_scalar` or :py:func:`persistence_scalar_index`.
Users of day ahead persistence forecasts will typically want to use
:py:func:`persistence_interval`.
:py:func:`persistence_interval_index`?

The functions accept a *load_data* keyword argument that allows users to
change where the functions load the observation data from. This is most
useful for users that would like to provide their own observation data
rather than using the solarforecastarbiter database.
"""
from functools import partial

import numpy as np
import pandas as pd

from solarforecastarbiter import datamodel, pvmodel

from statsmodels.distributions.empirical_distribution import ECDF


def persistence_scalar(observation, data_start, data_end, forecast_start,
                       forecast_end, interval_length, interval_label,
                       load_data):
    r"""
    Make a persistence forecast using the mean value of the
    *observation* from *data_start* to *data_end*.

    In the example below, we use GHI to be concrete but the concept
    applies to any kind of observation data. The persistence forecast
    is:

    .. math::

       GHI_{t_f} = \overline{GHI_{t_{start}} \ldots GHI_{t_{end}}}

    where :math:`t_f` is a forecast time, and the overline represents
    the average of all observations that occur between
    :math:`t_{start}` = *data_start* and :math:`t_{end}` = *data_end*.

    Parameters
    ----------
    observation : datamodel.Observation
    data_start : pd.Timestamp
        Observation data start. Forecast is inclusive of this instant if
        observation.interval_label is *beginning* or *instant*.
    data_end : pd.Timestamp
        Observation data end. Forecast is inclusive of this instant if
        observation.interval_label is *ending* or *instant*.
    forecast_start : pd.Timestamp
        Forecast start. Forecast is inclusive of this instant if
        interval_label is *beginning* or *instant*.
    forecast_end : pd.Timestamp
        Forecast end. Forecast is inclusive of this instant if
        interval_label is *ending* or *instant*.
    interval_length : pd.Timedelta
        Forecast interval length
    interval_label : str
        instant, beginning, or ending
    load_data : function
        A function that loads the observation data. Must have the
        signature load_data(observation, data_start, data_end) and
        properly account for observation interval label.

    Returns
    -------
    forecast : pd.Series
        The persistence forecast.
    """
    obs = load_data(observation, data_start, data_end)
    persistence_quantity = obs.mean()
    closed = datamodel.CLOSED_MAPPING[interval_label]
    fx_index = pd.date_range(start=forecast_start, end=forecast_end,
                             freq=interval_length, closed=closed)
    fx = pd.Series(persistence_quantity, index=fx_index)
    return fx


# Different signature than the scalar functions because forecast_end is
# determined by the forecast start and data length. Could make them the
# same for function call consistency, but readability counts.
def persistence_interval(observation, data_start, data_end, forecast_start,
                         interval_length, interval_label, load_data):
    r"""
    Make a persistence forecast for an *observation* using the mean
    values of each *interval_length* bin from *data_start* to
    *data_end*. The forecast starts at *forecast_start* and is of length
    *data_end* - *data_start*. A frequent use of this function is to
    create a day ahead persistence forecast using the previous day's
    observations.

    In the example below, we use GHI to be concrete but the concept
    applies to any kind of observation data. The persistence forecast
    for multiple intervals is:

    .. math::

       GHI_{t_{f_m}} = \overline{GHI_{t_{{start}_m}} \ldots GHI_{t_{{end}_m}}}

    where:

    .. math::

       m &\in \{0, 1, \ldots \frac{\textrm{data end} - \textrm{data start}}
           {\textrm{interval_length}} - 1\} \\
       t_{start_m} &=  \textrm{data start} +
           m \times \textrm{interval_length} \\
       t_{end_m} &= \textrm{data start} +
           (1 + m) \times \textrm{interval_length} \\
       t_{f_m} &= \textrm{forecast start} +
           m \times \textrm{interval_length} \\

    Further, persistence of multiple intervals requires that
    *data_start*, *data_end*, and *forecast_start* are all integer
    multiples of *interval_length*. For example, if *interval_length* =
    60 minutes, *data_start* may be 12:00 or 01:00, but not 12:30.

    Parameters
    ----------
    observation : datamodel.Observation
    data_start : pd.Timestamp
        Observation data start. Forecast is inclusive of this instant if
        observation.interval_label is *beginning* or *instant*.
    data_end : pd.Timestamp
        Observation data end. Forecast is inclusive of this instant if
        observation.interval_label is *ending* or *instant*.
    forecast_start : pd.Timestamp
        Forecast start. Forecast is inclusive of this instant if
        interval_label is *beginning* or *instant*.
    interval_length : pd.Timedelta
        Forecast interval length
    interval_label : str
        instant, beginning, or ending
    load_data : function
        A function that loads the observation data. Must have the
        signature load_data(observation, data_start, data_end) and
        properly account for observation interval label.

    Returns
    -------
    forecast : pd.Series
        The persistence forecast.
    """
    # determine forecast end from forecast start and obs length
    # assumes that load_data will return NaNs if data is missing.
    forecast_end = forecast_start + (data_end - data_start)

    # ensure that we're using times rounded to multiple of interval_length
    _check_intervals_times(observation.interval_label, data_start, data_end,
                           forecast_start, forecast_end,
                           observation.interval_length)

    closed_obs = datamodel.CLOSED_MAPPING[observation.interval_label]
    # get the data
    obs = load_data(observation, data_start, data_end)
    obs_index = pd.date_range(start=data_start, end=data_end,
                              freq=observation.interval_length,
                              closed=closed_obs)
    # put in nans if appropriate
    obs = obs.reindex(obs_index)
    # average data within bins of length interval_length
    persistence_quantity = obs.resample(interval_length,
                                        closed=closed_obs).mean()

    # Make the forecast time index.
    closed = datamodel.CLOSED_MAPPING[interval_label]
    fx_index = pd.date_range(start=forecast_start, end=forecast_end,
                             freq=interval_length, closed=closed)

    # Construct the returned series.
    # Use values to strip the time information from resampled obs.
    # Raises ValueError if len(persistence_quantity) != len(fx_index), but
    # that should never happen given the way we're calculating things here
    # now that nans are insterted into obs for missing data.
    fx = pd.Series(persistence_quantity.values, index=fx_index)
    return fx


def persistence_scalar_index(observation, data_start, data_end, forecast_start,
                             forecast_end, interval_length, interval_label,
                             load_data):
    r"""
    Calculate a persistence forecast using the mean value of the
    *observation* clear sky index or AC power index from *data_start* to
    *data_end*.

    In the example below, we use GHI to be concrete but the concept also
    applies to AC power. The persistence forecast is:

    .. math::

       GHI_{t_f} = \overline{
           \frac{ GHI_{t_{start}} }{ GHI_{{clear}_{t_{start}}} } \ldots
           \frac{ GHI_{t_{end}} }{ GHI_{{clear}_{t_{end}}} } }

    where :math:`t_f` is a forecast time, and the overline represents
    the average of all observations or clear sky values that occur
    between :math:`t_{start}` = *data_start* and
    :math:`t_{end}` = *data_end*. All :math:`GHI_{t}/GHI_{{clear}_t}`
    ratios are restricted to the range [0, 2] before the average is
    computed.

    Parameters
    ----------
    observation : datamodel.Observation
    data_start : pd.Timestamp
        Observation data start. Forecast is inclusive of this instant if
        observation.interval_label is *beginning* or *instant*.
    data_end : pd.Timestamp
        Observation data end. Forecast is inclusive of this instant if
        observation.interval_label is *ending* or *instant*.
    forecast_start : pd.Timestamp
        Forecast start. Forecast is inclusive of this instant if
        interval_label is *beginning* or *instant*.
    forecast_end : pd.Timestamp
        Forecast end. Forecast is inclusive of this instant if
        interval_label is *ending* or *instant*.
    interval_length : pd.Timedelta
        Forecast interval length
    interval_label : str
        instant, beginning, or ending
    load_data : function
        A function that loads the observation data. Must have the
        signature load_data(observation, data_start, data_end) and
        properly account for observation interval label.

    Returns
    -------
    forecast : pd.Series
        The persistence forecast. The forecast interval label is the
        same as the observation interval label.
    """
    # ensure that we're using times rounded to multiple of interval_length
    _check_intervals_times(observation.interval_label, data_start, data_end,
                           forecast_start, forecast_end,
                           observation.interval_length)

    # get observation data for specified range
    obs = load_data(observation, data_start, data_end)

    # partial-up the metadata for solar position and
    # clearsky calculation clarity and consistency
    site = observation.site
    calc_solpos = partial(pvmodel.calculate_solar_position,
                          site.latitude, site.longitude, site.elevation)
    calc_cs = partial(pvmodel.calculate_clearsky,
                      site.latitude, site.longitude, site.elevation)

    # Calculate solar position and clearsky for obs time range.
    # if data is instantaneous, calculate at the obs time.
    # else (if data is interval average), calculate at 1 minute resolution to
    # reduce errors from changing solar position during persistence data range.
    # Later, modeled clear sky or ac power will be averaged over the data range
    closed = datamodel.CLOSED_MAPPING[observation.interval_label]
    if closed is None:
        freq = observation.interval_length
    else:
        freq = pd.Timedelta('1min')
    obs_range = pd.date_range(start=data_start, end=data_end, freq=freq,
                              closed=closed)
    solar_position_obs = calc_solpos(obs_range)
    clearsky_obs = calc_cs(solar_position_obs['apparent_zenith'])

    # Calculate solar position and clearsky for the forecast times.
    # Use 5 minute or better frequency to minimize solar position errors.
    # Later, modeled clear sky or ac power will be resampled to interval_length
    closed_fx = datamodel.CLOSED_MAPPING[interval_label]
    fx_range = pd.date_range(start=forecast_start, end=forecast_end, freq=freq,
                             closed=closed_fx)
    solar_position_fx = calc_solpos(fx_range)
    clearsky_fx = calc_cs(solar_position_fx['apparent_zenith'])

    # Consider putting the code within each if/else block below into its own
    # function with standard outputs clear_ref and clear_fx. But with only two
    # cases for now, it might be more clear to leave inline.
    if (
            isinstance(site, datamodel.SolarPowerPlant) and
            observation.variable == 'ac_power'
    ):
        # No temperature input is only OK so long as temperature effects
        # do not push the system above or below AC clip point.
        # It's only a reference forecast!
        clear_ref = pvmodel.irradiance_to_power(
            site.modeling_parameters, solar_position_obs['apparent_zenith'],
            solar_position_obs['azimuth'], clearsky_obs['ghi'],
            clearsky_obs['dni'], clearsky_obs['dhi']
        )
        clear_fx = pvmodel.irradiance_to_power(
            site.modeling_parameters, solar_position_fx['apparent_zenith'],
            solar_position_fx['azimuth'], clearsky_fx['ghi'],
            clearsky_fx['dni'], clearsky_fx['dhi']
        )
    else:
        # assume we are working with ghi, dni, or dhi.
        clear_ref = clearsky_obs[observation.variable]
        clear_fx = clearsky_fx[observation.variable]

    # resample sub-interval reference clear sky to observation intervals
    clear_ref_resampled = clear_ref.resample(
        observation.interval_length, closed=closed, label=closed).mean()
    # calculate persistence index (clear sky index or ac power index)
    # avg{index_{t_start}...index_{t_end}} =
    #   avg{obs_{t_start}/clear_{t_start}...obs_{t_end}/clear_{t_end}}
    # clear_ref is calculated at high temporal resolution, so this is accurate
    # for any observation interval length
    # apply clip to the clear sky index array before computing the average.
    # this prevents outliers from distorting the mean, a common occurance
    # near sunrise and sunset.
    pers_index = (obs / clear_ref_resampled).clip(lower=0, upper=2).mean()

    # average instantaneous clear forecasts over interval_length windows
    # resample operation should be safe due to
    # _check_interval_length calls above
    clear_fx_resampled = clear_fx.resample(
        interval_length, closed=closed_fx, label=closed_fx).mean()

    # finally, make forecast
    # fx_t_f = avg{index_{t_start}...index_{t_end}} * clear_t_f
    fx = pers_index * clear_fx_resampled

    return fx


def _check_intervals_times(interval_label, data_start, data_end,
                           forecast_start, forecast_end, interval_length):
    """Ensures valid input.

    Raises
    ------
    ValueError
    """
    interval_length_sec = int(interval_length.total_seconds())
    data_start_mod = data_start.timestamp() % interval_length_sec == 0
    forecast_start_mod = forecast_start.timestamp() % interval_length_sec == 0
    data_end_mod = data_end.timestamp() % interval_length_sec == 0
    forecast_end_mod = forecast_end.timestamp() % interval_length_sec == 0
    strvals = (
        f'interval_label={interval_label}, data_start={data_start}, '
        f'data_end={data_end}, forecast_start={forecast_start}, '
        f'forecast_end={forecast_end}, interval_length={interval_length_sec}s')
    if 'instant' in interval_label:
        # two options allowed:
        # 1. start times % int length == 0 and NOT end times % int. length == 0
        # 2. NOT start times % int length == 0 and end times % int. length == 0
        # everything else fails
        if data_start_mod and not data_end_mod:
            pass
        elif not data_start_mod and data_end_mod:
            pass
        else:
            raise ValueError('For observations with interval_label '
                             'instant, data_start OR data_end '
                             'must be must be divisible by interval_length.' +
                             strvals)
    elif interval_label in ['ending', 'beginning']:
        if not all((data_start_mod, forecast_start_mod,
                    data_end_mod, forecast_end_mod)):
            raise ValueError('For observations with interval_label beginning '
                             'or ending, all of data_start, forecast_start, '
                             'data_end, and forecast_end must be divisible by '
                             'interval_length. ' + strvals)
    else:
        raise ValueError('invalid interval_label')


def persistence_probabilistic(observation, data_start, data_end,
                              forecast_start, forecast_end,
                              interval_length, interval_label,
                              load_data, axis, constant_values):
    r"""
    Make a probabilistic persistence forecast using the *observation* from
    *data_start* to *data_end*. In the forecast literature, this method is
    typically referred to as Persistence Ensemble (PeEn). [1]_ [2]_ [3]_

    The function handles forecasting either constant variable values or
    constant percentiles. In the examples below, we use GHI to be concrete but
    the concepts also apply to other variables (AC power, net load, etc.).

    If forecasting constant variable values (e.g. forecast the probability of
    GHI being less than or equal to 500 W/m^2), the persistence forecast is:

    .. math::

       F_n(x) = ECDF(GHI_{t_{start}}, ..., GHI_{t_{end}})
       Prob(GHI_{t_f} <= 100 W/m^2) = F_n(100 W/m^2)

    where :math:`t_f` is a forecast time and :math:`F_n` is the empirical CDF
    (ECDF) function computed from the *n* observations between
    :math:`t_{start}` = *data_start* and :math:`t_{end}` = *data_end*, which
    maps from variable values to probabilities.

    If forecasting constant probabilities (e.g. forecast the GHI value that has
    a 50% probability), the persistence forecast is:

    .. math::

       F_n(x) = ECDF(GHI_{t_{start}}, ..., GHI_{t_{end}})
       Q_n(p) = \inf {x \in \mathrf{R} : p \leq F_n(x) }
       p_{t_f} = Q_n(50%)

    where :math:`Q_n` is the quantile function based on the *n* observations
    between :math:`t_{start}` = *data_start* and :math:`t_{end}` = *data_end*,
    which maps from probabilities to variable values.

    Parameters
    ----------
    observation : datamodel.Observation
    data_start : pd.Timestamp
        Observation data start. Forecast is inclusive of this instant if
        observation.interval_label is *beginning* or *instant*.
    data_end : pd.Timestamp
        Observation data end. Forecast is inclusive of this instant if
        observation.interval_label is *ending* or *instant*.
    forecast_start : pd.Timestamp
        Forecast start. Forecast is inclusive of this instant if
        interval_label is *beginning* or *instant*.
    forecast_end : pd.Timestamp
        Forecast end. Forecast is inclusive of this instant if
        interval_label is *ending* or *instant*.
    interval_length : pd.Timedelta
        Forecast interval length
    interval_label : str
        instant, beginning, or ending
    load_data : function
        A function that loads the observation data. Must have the
        signature load_data(observation, data_start, data_end) and
        properly account for observation interval label.
    axis : {'x', 'y'}
        The axis on which the constant values of the CDF is specified. The axis
        can be either *x* (constant variable values) or *y* (constant
        percentiles).
    constant_values : array_like
        The variable values or percentiles.

    Returns
    -------
    forecasts : list of pd.Series
        The persistence forecasts, returned in the same order as
        *constant_values*. If *axis* is *x*, the forecast values are
        percentiles (e.g. 25%). If instead *axis* is *y*, the forecasts values
        have the same units as the observation data (e.g. MW).

    Raises
    ------
    ValueError
        If the **axis** parameter is invalid.

    References
    ----------
    .. [1] Allessandrini et al. (2015) "An analog ensemble for short-term
       probabilistic solar power forecast", Appl. Energy 157, pp. 95-110.
       doi: 10.1016/j.apenergy.2015.08.011
    .. [2] Yang (2019) "A universal benchmarking method for probabilistic
       solar irradiance forecasting", Solar Energy 184, pp. 410-416.
       doi: 10.1016/j.solener.2019.04.018
    .. [3] Doubleday, Van Scyoc Herndandez and Hodge (2020) "Benchmark
       probabilistic solar forecasts: characteristics and recommendations",
       Solar Energy 206, pp. 52-67. doi: 10.1016/j.solener.2020.05.051

    """
    closed = datamodel.CLOSED_MAPPING[interval_label]
    fx_index = pd.date_range(start=forecast_start, end=forecast_end,
                             freq=interval_length, closed=closed)

    # observation data resampled to match forecast sampling
    obs = load_data(observation, data_start, data_end)
    obs = obs.resample(interval_length, closed=closed).mean()

    if obs.empty:
        return [pd.Series(None, dtype=float, index=fx_index)
                for _ in constant_values]

    if axis == "x":
        cdf = ECDF(obs)
        forecasts = []
        for constant_value in constant_values:
            fx_prob = cdf(constant_value) * 100.0
            forecasts.append(pd.Series(fx_prob, index=fx_index))
    elif axis == "y":   # constant_values=percentiles, fx=variable
        forecasts = []
        for constant_value in constant_values:
            fx = np.percentile(obs, constant_value)
            forecasts.append(pd.Series(fx, index=fx_index))
    else:
        raise ValueError(f"Invalid axis parameter: {axis}")

    return forecasts


def persistence_probabilistic_timeofday(observation, data_start, data_end,
                                        forecast_start, forecast_end,
                                        interval_length, interval_label,
                                        load_data, axis, constant_values):
    r"""
    Make a probabilistic persistence forecast using the *observation* from
    *data_start* to *data_end*, matched by time of day (e.g. to forecast 9am,
    only use observations from 9am on days between *data_start* and
    *data_end*). This is a common variant of the Persistence Ensemble (PeEn)
    method. [1]_ [2]_ [3]_

    Parameters
    ----------
    observation : datamodel.Observation
    data_start : pd.Timestamp
        Observation data start. Forecast is inclusive of this instant if
        observation.interval_label is *beginning* or *instant*.
    data_end : pd.Timestamp
        Observation data end. Forecast is inclusive of this instant if
        observation.interval_label is *ending* or *instant*.
    forecast_start : pd.Timestamp
        Forecast start. Forecast is inclusive of this instant if
        interval_label is *beginning* or *instant*.
    forecast_end : pd.Timestamp
        Forecast end. Forecast is inclusive of this instant if
        interval_label is *ending* or *instant*.
    interval_length : pd.Timedelta
        Forecast interval length
    interval_label : str
        instant, beginning, or ending
    load_data : function
        A function that loads the observation data. Must have the
        signature load_data(observation, data_start, data_end) and
        properly account for observation interval label.
    axis : {'x', 'y'}
        The axis on which the constant values of the CDF is specified. The axis
        can be either *x* (constant variable values) or *y* (constant
        percentiles).
    constant_values : array_like
        The variable values or percentiles.

    Returns
    -------
    forecasts : list of pd.Series
        The persistence forecasts, returned in the same order as
        *constant_values*. If *axis* is *x*, the forecast values are
        percentiles (e.g. 25%). If instead *axis* is *y*, the forecasts values
        have the same units as the observation data (e.g. MW).

    Raises
    ------
    ValueError
        If there is insufficient data for matching by time of day or the
        **axis** parameter is invalid.

    Notes
    -----
    Assumes that there is at least 20 days of *observation* data available
    based on [1]_, [2]_, [3]_.

    References
    ----------
    .. [1] Allessandrini et al. (2015) "An analog ensemble for short-term
       probabilistic solar power forecast", Appl. Energy 157, pp. 95-110.
       doi: 10.1016/j.apenergy.2015.08.011
    .. [2] Yang (2019) "A universal benchmarking method for probabilistic
       solar irradiance forecasting", Solar Energy 184, pp. 410-416.
       doi: 10.1016/j.solener.2019.04.018
    .. [3] Doubleday, Van Scyoc Herndandez and Hodge (2020) "Benchmark
       probabilistic solar forecasts: characteristics and recommendations",
       Solar Energy 206, pp. 52-67. doi: 10.1016/j.solener.2020.05.051

    See also
    --------
    :py:func:`solarforecastarbiter.reference_forecasts.persistence.persistence_probabilistic`

    """
    # ensure that we're using times rounded to multiple of interval_length
    _check_intervals_times(observation.interval_label, data_start, data_end,
                           forecast_start, forecast_end,
                           observation.interval_length)

    closed = datamodel.CLOSED_MAPPING[interval_label]
    fx_index = pd.date_range(start=forecast_start, end=forecast_end,
                             freq=interval_length, closed=closed)

    # observation data resampled to match forecast sampling
    obs = load_data(observation, data_start, data_end)
    obs = obs.resample(interval_length, closed=closed).mean()
    if obs.empty:
        raise ValueError("Insufficient data to match by time of day")

    # time of day: minutes past midnight (e.g. 0=12:00am, 75=1:15am)
    if obs.index.tzinfo is not None:
        if fx_index.tzinfo is not None:
            obs = obs.tz_convert(fx_index.tzinfo)
        else:
            fx_index = fx_index.tz_localize(obs.index.tzinfo)
    else:
        if fx_index.tzinfo is not None:
            obs = obs.tz_localize(fx_index.tzinfo)

    obs_timeofday = (obs.index.hour * 60 + obs.index.minute).astype(int)
    fx_timeofday = (fx_index.hour * 60 + fx_index.minute).astype(int)

    # confirm sufficient data for matching by time of day
    if obs.last_valid_index() - obs.first_valid_index() < pd.Timedelta("20D"):
        raise ValueError("Insufficient data to match by time of day")

    if axis == "x":
        forecasts = []
        for constant_value in constant_values:
            fx = pd.Series(np.nan, index=fx_index)
            for tod in fx_timeofday:
                data = obs[obs_timeofday == tod]
                cdf = ECDF(data)
                fx[fx_timeofday == tod] = cdf(constant_value) * 100.0
            forecasts.append(fx)
    elif axis == "y":
        forecasts = []
        for constant_value in constant_values:
            fx = pd.Series(np.nan, index=fx_index)
            for tod in fx_timeofday:
                data = obs[obs_timeofday == tod]
                fx[fx_timeofday == tod] = np.percentile(data, constant_value)
            forecasts.append(fx)
    else:
        raise ValueError(f"Invalid axis parameter: {axis}")

    return forecasts



from functools import partial
import logging
import os


import pandas as pd
from requests.exceptions import HTTPError


from solarforecastarbiter.io.fetch import rtc
from solarforecastarbiter.io.reference_observations import (
    common, default_forecasts)


logger = logging.getLogger('reference_data')


DOE_RTC_VARIABLE_MAP = {
    'GlobalIrrad': 'ghi',
    'DiffuseIrrad': 'dhi',
    'DirectIrrad': 'dni',
    'AmbientTemp_weather': 'air_temperature',
    'WindSpeed': 'wind_speed',
    'RelativeHumidity': 'relative_humidity',
    'PyranometerIrrad': 'poa_global',
    'Sys1Wac': 'ac_power'
}


def adjust_site_parameters(site):
    """Add the PV modeling parameters"""
    out = site.copy()
    modeling_params = {
        'ac_capacity': 0.00324,  # no clipping
        'dc_capacity': 0.00324,
        'temperature_coefficient': -0.420,
        'dc_loss_factor': 0,
        'ac_loss_factor': 0,
        'surface_tilt': 35,
        'surface_azimuth': 180,
        'tracking_type': 'fixed'}
    out['modeling_parameters'] = modeling_params
    out['extra_parameters']['module'] = 'Suniva 270W'
    return out


def initialize_site_observations(api, site):
    """Creates an observation at the site for each variable in
    DOE_RTC_VARIABLE_MAP

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An active Reference user session.
    site : datamodel.Site
        The site object for which to create Observations.
    """
    for sfa_var in DOE_RTC_VARIABLE_MAP.values():
        logger.info(f'Creating {sfa_var} at {site.name}')
        try:
            common.create_observation(
                api, site, sfa_var)
        except HTTPError as e:
            logger.error(f'Could not create Observation for "{sfa_var}" '
                         f'at DOE RTC site {site.name}')
            logger.debug(f'Error: {e.response.text}')


def initialize_site_forecasts(api, site):
    """
    Create a forecasts for each variable in DOE_RTC_VARIABLE_MAP at each site

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An active Reference user session.
    site : datamodel.Site
        The site object for which to create Forecasts.
    """
    common.create_forecasts(api, site, DOE_RTC_VARIABLE_MAP.values(),
                            default_forecasts.TEMPLATE_FORECASTS)


def fetch(api, site, start, end, *, doe_rtc_api_key):
    """Retrieve observation data for a DOE RTC site between start and end.

    Parameters
    ----------
    api : io.APISession
        Unused but conforms to common.update_site_observations call
    site : datamodel.Site
        Site object with the appropriate metadata.
    start : datetime
        The beginning of the period to request data for.
    end : datetime
        The end of the period to request data for.
    doe_rtc_api_key : str
        API key to access the DOE RTC api

    Returns
    -------
    data : pandas.DataFrame
        All of the requested data concatenated into a single DataFrame.
    """
    try:
        site_extra_params = common.decode_extra_parameters(site)
    except ValueError:
        return pd.DataFrame()
    doe_rtc_site_id = site_extra_params['network_api_id']
    obs_df = rtc.fetch_doe_rtc(
        doe_rtc_site_id, doe_rtc_api_key,
        start.tz_convert(site.timezone), end.tz_convert(site.timezone))
    if obs_df.empty:
        logger.warning(f'Data for site {site.name} contained no '
                       f'entries from {start} to {end}.')
        return pd.DataFrame()
    obs_df = obs_df.rename(columns=DOE_RTC_VARIABLE_MAP).tz_localize(
        site.timezone)
    # W to MW
    if 'ac_power' in obs_df:
        obs_df['ac_power'] = obs_df['ac_power'] / 1e6
    return obs_df


def update_observation_data(api, sites, observations, start, end):
    """Post new observation data to a list of DOE RTC Observations
    from start to end.

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An active Reference user session.
    sites: list of solarforecastarbiter.datamodel.Site
        List of all reference sites as Objects
    observations: list of solarforecastarbiter.datamodel.Observation
        List of all reference observations.
    start : datetime
        The beginning of the period to request data for.
    end : datetime
        The end of the period to request data for.
    """
    doe_rtc_api_key = os.getenv('DOE_RTC_API_KEY')
    if doe_rtc_api_key is None:
        raise KeyError('"DOE_RTC_API_KEY" environment variable must be '
                       'set to update DOE RTC observation data.')
    doe_rtc_sites = common.filter_by_networks(sites, 'DOE RTC')
    for site in doe_rtc_sites:
        common.update_site_observations(
            api, partial(fetch, doe_rtc_api_key=doe_rtc_api_key), site,
            observations, start, end)



from functools import partial
import json
import logging
import os
from pkg_resources import resource_filename, Requirement


import pandas as pd
from requests.exceptions import HTTPError


from solarforecastarbiter.io.fetch import arm
from solarforecastarbiter.io.reference_observations import (
    common, default_forecasts)


DEFAULT_SITEFILE = resource_filename(
    Requirement.parse('solarforecastarbiter'),
    'solarforecastarbiter/io/reference_observations/'
    'arm_reference_sites.json')

DOE_ARM_SITE_VARIABLES = {
    'qcrad': arm.IRRAD_VARIABLES,
    'met': arm.MET_VARIABLES,
}

DOE_ARM_VARIABLE_MAP = {
    'down_short_hemisp': 'ghi',
    'short_direct_normal': 'dni',
    'down_short_diffuse_hemisp': 'dhi',
    'temp_mean': 'air_temperature',
    'rh_mean': 'relative_humidity',
    'wspd_arith_mean': 'wind_speed',
}

logger = logging.getLogger('reference_data')


def _determine_stream_vars(datastream):
    """Returns a list of variables available based on datastream name.

    Parameters
    ----------
    datastream: str
        Datastream name, or the product name. This string is searched for
        `met` or `qcrad` and returns a list of expected variables.

    Returns
    -------
    list of str
        The variable names that can be found in the file.
    """
    available = []
    for stream_type, arm_vars in DOE_ARM_SITE_VARIABLES.items():
        if stream_type in datastream:
            available = available + arm_vars
    return available


def initialize_site_observations(api, site):
    """Creates an observation at the site for each variable in
    the matched DOE_ARM_VARIABLE_MAP.

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An active Reference user session.
    site : datamodel.Site
        The site object for which to create Observations.
    """
    try:
        site_extra_params = common.decode_extra_parameters(site)
    except ValueError:
        logger.error(f'Failed to initialize observations  for {site.name} '
                     'extra parameters could not be loaded.')
        return

    site_vars = site_variables_from_extra_params(site_extra_params)
    for sfa_var in site_vars:
        logger.info(f'Creating {sfa_var} at {site.name}')
        try:
            common.create_observation(
                api, site, sfa_var)
        except HTTPError as e:
            logger.error(f'Could not create Observation for "{sfa_var}" '
                         f'at DOE ARM site {site.name}')
            logger.debug(f'Error: {e.response.text}')


def initialize_site_forecasts(api, site):
    """
    Create a forecast for each variable at the site.

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An active Reference user session.
    site : datamodel.Site
        The site object for which to create Forecasts.
    """
    try:
        site_extra_params = common.decode_extra_parameters(site)
    except ValueError:
        logger.error('Failed to initialize reference forecasts for '
                     f'{site.name} extra parameters could not be loaded.')
        return

    site_vars = site_variables_from_extra_params(site_extra_params)

    common.create_forecasts(api, site, site_vars,
                            default_forecasts.TEMPLATE_FORECASTS)


def fetch(api, site, start, end, *, doe_arm_user_id, doe_arm_api_key):
    """Retrieve observation data for a DOE ARM site between start and end.

    Parameters
    ----------
    api : io.APISession
        Unused but conforms to common.update_site_observations call
    site : datamodel.Site
        Site object with the appropriate metadata.
    start : datetime
        The beginning of the period to request data for.
    end : datetime
        The end of the period to request data for.
    doe_arm_user_id : str
        User ID to access the DOE ARM api.
    doe_arm_api_key : str
        API key to access the DOE ARM api.

    Returns
    -------
    data : pandas.DataFrame
        All of the requested data concatenated into a single DataFrame.
    """
    try:
        site_extra_params = common.decode_extra_parameters(site)
    except ValueError:
        return pd.DataFrame()

    available_datastreams = site_extra_params['datastreams']

    datastreams = {}
    # Build a dict with top-level keys to 'met' and 'qcrad' if meteorological
    # or irradiance  data exists at the site. This is to later group dataframes
    # created from each datastream by the type of data found in the stream.
    for ds_type in ['met', 'qcrad']:
        if ds_type in available_datastreams:
            ds_type_dict = {}
            streams = available_datastreams[ds_type]

            # When a dict is present each key is a datastream and value is
            # a date range for which the datastream contains data. We need to
            # determine which streams to use to get all of the requested data.
            if isinstance(streams, dict):
                ds_type_dict.update(
                    find_stream_data_availability(streams, start, end))
            else:
                # If a single string datastream name exists, we assume that all
                # available data is contained in the stream. Deferring to the
                # data fetch process, which will fail to retrieve data and
                # continue gracefully.
                ds_type_dict[streams] = (start, end)
            datastreams[ds_type] = ds_type_dict

    site_dfs = []

    for stream_type in datastreams:
        # Stitch together all the datastreams with similar data.
        stream_type_dfs = []
        for datastream, date_range in datastreams[stream_type].items():
            stream_df = arm.fetch_arm(
                doe_arm_user_id,
                doe_arm_api_key,
                datastream,
                _determine_stream_vars(datastream),
                date_range[0].tz_convert(site.timezone),
                date_range[1].tz_convert(site.timezone)
            )
            if stream_df.empty:
                logger.warning(f'Datastream {datastream} for site {site.name} '
                               f'contained no entries from {start} to {end}.')
            else:
                stream_type_dfs.append(stream_df)
        if stream_type_dfs:
            # Concatenate all dataframes of similar data
            stream_type_df = pd.concat(stream_type_dfs)
            site_dfs.append(stream_type_df)

    if site_dfs:
        # Join dataframes with different variables along the index, this has
        # the side effect of introducing missing data if any requests have
        # failed.
        obs_df = pd.concat(site_dfs, axis=1)
        obs_df = obs_df.rename(columns=DOE_ARM_VARIABLE_MAP)
        return obs_df
    else:
        logger.warning(f'Data for site {site.name} contained no entries from '
                       f'{start} to {end}.')
        return pd.DataFrame()


def update_observation_data(api, sites, observations, start, end):
    """Post new observation data to a list of DOE ARM Observations
    from start to end.

    api : solarforecastarbiter.io.api.APISession
        An active Reference user session.
    sites: list of solarforecastarbiter.datamodel.Site
        List of all reference sites as Objects
    observations: list of solarforecastarbiter.datamodel.Observation
        List of all reference observations.
    start : datetime
        The beginning of the period to request data for.
    end : datetime
        The end of the period to request data for.
    """
    doe_arm_api_key = os.getenv('DOE_ARM_API_KEY')
    if doe_arm_api_key is None:
        raise KeyError('"DOE_ARM_API_KEY" environment variable must be '
                       'set to update DOE ARM observation data.')
    doe_arm_user_id = os.getenv('DOE_ARM_USER_ID')
    if doe_arm_user_id is None:
        raise KeyError('"DOE_ARM_USER_ID" environment variable must be '
                       'set to update DOE ARM observation data.')

    doe_arm_sites = common.filter_by_networks(sites, 'DOE ARM')
    for site in doe_arm_sites:
        common.update_site_observations(
            api, partial(fetch, doe_arm_user_id=doe_arm_user_id,
                         doe_arm_api_key=doe_arm_api_key),
            site, observations, start, end)


def adjust_site_parameters(site):
    """Updates extra parameters with applicable datastreams from
    `arm_reference_sites.json`

    Parameters
    ----------
    site: dict

    Returns
    -------
    dict
        Copy of input with updated extra parameters.
    """
    with open(DEFAULT_SITEFILE) as fp:
        sites_metadata = json.load(fp)['sites']

    # ARM has multiple 'locations' at each 'site'. For example, the Southern
    # Great Plains (SGP) ARM site has many locations throughout Oklahoma, and
    # neighboring states. Each location is identified by a code, e.g. Byron
    # Oklahoma is location `E11` at the SGP site, and it's data is accessed via
    # datastreams with the pattern `sgp<product>e11.<data-level>` where product
    # indicates the contents of the datastream (we are interested in `met` and
    # `qcrad1long` products) and data-level indicates quality and any
    # processing applied to the data. In the Solar Forecast Arbiter we store
    # each 'location' as a SFA site. We use the `network_api_id` to indicate
    # the ARM site's location code and `network_api_abbreviation` to represent
    # ARM "site code (e.g. `sgp`). Both of these keys must be used to  identify
    # a site in the Solar Forecast Arbiter, because ARM's location ids are only
    # unique for a given ARM site.
    arm_location_id = site['extra_parameters']['network_api_id']
    arm_site_id = site['extra_parameters']['network_api_abbreviation']

    for site_metadata in sites_metadata:
        site_extra_params = json.loads(site_metadata['extra_parameters'])
        if (
            site_extra_params['network_api_id'] == arm_location_id
            and site_extra_params['network_api_abbreviation'] == arm_site_id

        ):
            site_out = site.copy()
            site_out['extra_parameters'] = site_extra_params
            return site_out
    return site


def find_stream_data_availability(streams, start, end):
    """Determines what date ranges to use for each datastream. Date ranges of
    each stream must not overlap.

    Parameters
    ----------
    streams: dict
        Dict where values are string datastream names and values are iso8601
        date ranges `start/end` indicating the period of data available at
        that datastream.
    start: datetime
        The start of the period to request data for.
    end: datetime
        The end of the period to request data for.

    Returns
    -------
    dict
        Dict where keys are datastreams and values are two element lists of
        `[start datetime, end datetime]` that when considered together should
        span all of the available data between the requested start and end.

    Raises
    ------
    ValueError
        If streams in the dictionary have overlapping ranges.
    """
    stream_range_dict = {}
    streams_with_ranges = [(stream, parse_iso_date_range(date_range))
                           for stream, date_range in streams.items()]
    streams_overlap = detect_stream_overlap(streams_with_ranges)
    if streams_overlap:
        raise ValueError(f'Overlapping datastreams found in {streams.keys()}.')

    # Find the overlap between each streams available data, and the requested
    # period
    for datastream, stream_range in streams_with_ranges:
        overlap = get_period_overlap(
            start, end, stream_range[0], stream_range[1])
        if overlap is None:
            # The datastream did not contain any data within the requested
            # range, we don't need to use it for this request.
            continue
        else:
            stream_range_dict[datastream] = overlap
    return stream_range_dict


def get_period_overlap(request_start, request_end, avail_start, avail_end):
    """Finds period of overlap between the requested time range and the
    available period.

    Parameters
    ----------
    request_start: datetime-like
        Start of the period of requested data.
    request_end: datetime-like
        End of the period of requested data.
    avail_start: datetime-like
        Start of the available data.
    avail_end: datatime-like
        End of available data.

    Returns
    -------
    start, end: list of datetime or None
        Start and end of overlapping period, or None if no overlap occurred.
    """
    if request_start < avail_end and request_end > avail_start:
        if request_start < avail_start:
            start = avail_start
        else:
            start = request_start
        if request_end > avail_end:
            end = avail_end
        else:
            end = request_end
        return [start, end]
    else:
        return None


def parse_iso_date_range(date_range_string):
    """Parses a date range string in iso8601 format ("start date/end date")
    into a tuple of pandas timestamps `(start, end)`.
    """
    start, end = date_range_string.split('/')
    return (pd.Timestamp(start, tz='utc'), pd.Timestamp(end, tz='utc'))


def site_variables_from_extra_params(site_extra_params):
    """Return variables expected at the site, based on the content of the
    `datastreams` attribute of the site's extra parameters.
    """
    return [DOE_ARM_VARIABLE_MAP[arm_var]
            for stream_type in site_extra_params['datastreams'].keys()
            for arm_var in _determine_stream_vars(stream_type)]


def detect_stream_overlap(stream_list):
    """Detects if a list of streams contain any overlap

    Parameters
    ----------
    stream_list: list of 3-tuples
        List of (datastream, (start, end)) where datastream is a string, start
        and end are pandas timestamps.

    Returns
    -------
    bool
        Returns True if any of the stream's ranges overlap,  otherwise False.
    """
    overlap = []
    for i, stream in enumerate(stream_list):
        date_range = stream[1]
        for j, other_stream in enumerate(stream_list[:i] + stream_list[i+1:]):
            other_date_range = other_stream[1]
            overlap.append(get_period_overlap(
                date_range[0], date_range[1],
                other_date_range[0], other_date_range[1]))
    return any(overlap)



from pathlib import Path
import re
import uuid


import numpy as np
import pandas as pd
import pytest


from solarforecastarbiter import datamodel
from solarforecastarbiter.io import api
from solarforecastarbiter.reports import main


EMPTY_DF = pd.DataFrame(columns=['value', 'quality_flag'],
                        index=pd.DatetimeIndex([], tz='UTC'))


@pytest.fixture()
def _test_data(report_objects, ref_forecast_id, remove_orca):
    report, observation, forecast_0, forecast_1, aggregate, forecast_agg = \
        report_objects
    base = Path(__file__).resolve().parent
    data = {
        observation.observation_id: pd.read_csv(
            base / 'observation_values.csv', index_col='timestamp',
            parse_dates=True),
        forecast_0.forecast_id: pd.read_csv(
            base / 'forecast_0_values.csv', header=None, parse_dates=True,
            names=['timestamp', 'value'], index_col='timestamp')['value'],
        forecast_1.forecast_id: pd.read_csv(
            base / 'forecast_1_values.csv', header=None, parse_dates=True,
            names=['timestamp', 'value'], index_col='timestamp')['value'],
        aggregate.aggregate_id: pd.read_csv(
            base / 'observation_values.csv', index_col='timestamp',
            parse_dates=True),
        forecast_agg.forecast_id: pd.read_csv(
            base / 'forecast_0_values.csv', header=None, parse_dates=True,
            names=['timestamp', 'value'], index_col='timestamp')['value'],
    }
    data[ref_forecast_id] = data[forecast_0.forecast_id] + 1
    return data


@pytest.fixture()
def mock_data(mocker, _test_data):
    def get_data(id_, start, end, interval_label=None):
        return _test_data[id_].loc[start:end]

    get_forecast_values = mocker.patch(
        'solarforecastarbiter.reports.main.APISession.get_forecast_values',
        side_effect=get_data)
    get_observation_values = mocker.patch(
        'solarforecastarbiter.reports.main.APISession.get_observation_values',
        side_effect=get_data)
    get_aggregate_values = mocker.patch(
        'solarforecastarbiter.reports.main.APISession.get_aggregate_values',
        side_effect=get_data)

    return get_forecast_values, get_observation_values, get_aggregate_values


def test_get_data_for_report(mock_data, report_objects, mocker):
    report, observation, forecast_0, forecast_1, aggregate, forecast_agg = \
        report_objects
    session = api.APISession('nope')
    apply_obs = mocker.spy(main, 'apply_validation')
    data = main.get_data_for_report(session, report)
    assert apply_obs.call_count == 2
    assert isinstance(data[observation], pd.DataFrame)
    assert isinstance(data[forecast_0], pd.Series)
    assert isinstance(data[forecast_1], pd.Series)
    assert isinstance(data[aggregate], pd.DataFrame)
    assert isinstance(data[forecast_agg], pd.Series)
    get_forecast_values, get_observation_values, get_aggregate_values = \
        mock_data
    assert get_forecast_values.call_count == 4
    assert get_observation_values.call_count == 1
    assert get_aggregate_values.call_count == 1


@pytest.fixture()
def _test_event_data(event_report_objects, remove_orca):
    report, observation, forecast_0, forecast_1 = event_report_objects

    tz = "US/Pacific"
    index = pd.date_range(start="20200401T0000", end="20200404T2359",
                          freq="1h", tz=tz)
    obs_series = pd.DataFrame(data={
        "value": np.random.randint(0, 2, len(index), dtype=bool),
        "quality_flag": 2
    }, index=index)

    fx0_series = pd.Series(
        index=index, data=np.random.randint(0, 2, len(index), dtype=bool)
    )

    fx1_series = pd.Series(
        index=index, data=np.random.randint(0, 2, len(index), dtype=bool)
    )

    data = {
        observation.observation_id: obs_series,
        forecast_0.forecast_id: fx0_series,
        forecast_1.forecast_id: fx1_series,
    }
    return data


@pytest.fixture()
def mock_event_data(mocker, _test_event_data):
    def get_data(id_, start, end, interval_label=None):
        return _test_event_data[id_].loc[start:end]

    get_forecast_values = mocker.patch(
        'solarforecastarbiter.reports.main.APISession.get_forecast_values',
        side_effect=get_data)
    get_observation_values = mocker.patch(
        'solarforecastarbiter.reports.main.APISession.get_observation_values',
        side_effect=get_data)

    return get_forecast_values, get_observation_values


def test_get_data_for_report_event(mock_event_data, event_report_objects):
    report, observation, forecast_0, forecast_1 = event_report_objects
    session = api.APISession('nope')
    data = main.get_data_for_report(session, report)
    assert isinstance(data[observation], pd.DataFrame)
    assert isinstance(data[forecast_0], pd.Series)
    assert isinstance(data[forecast_1], pd.Series)
    get_forecast_values, get_observation_values = mock_event_data
    assert get_forecast_values.call_count == 2
    assert get_observation_values.call_count == 1


def test_get_version():
    vers = main.get_versions()
    assert {v[0] for v in vers} > {'solarforecastarbiter', 'python'}


def test_infer_timezone(report_objects):
    report = report_objects[0]
    assert main.infer_timezone(report.report_parameters) == "Etc/GMT+7"


def test_infer_timezone_agg(report_objects, single_forecast_aggregate,
                            single_forecast_observation):
    report_params = report_objects[0].report_parameters
    rp = report_params.replace(object_pairs=(single_forecast_aggregate,
                                             single_forecast_observation))
    assert main.infer_timezone(rp) == 'America/Denver'


def test_create_raw_report_from_data(mocker, report_objects, _test_data):
    report = report_objects[0]
    data = {}
    for fxobs in report.report_parameters.object_pairs:
        data[fxobs.forecast] = _test_data[fxobs.forecast.forecast_id]
        if fxobs.reference_forecast is not None:
            data[fxobs.reference_forecast] = _test_data[
                fxobs.reference_forecast.forecast_id]
        if isinstance(fxobs, datamodel.ForecastAggregate):
            data[fxobs.aggregate] = _test_data[fxobs.aggregate.aggregate_id]
        else:
            data[fxobs.observation] = _test_data[
                fxobs.observation.observation_id]

    raw = main.create_raw_report_from_data(report, data)
    assert isinstance(raw, datamodel.RawReport)
    assert len(raw.plots.figures) > 0


def test_create_raw_report_from_data_no_fx(mocker, report_objects, _test_data):
    report = report_objects[0]
    data = {}
    for fxobs in report.report_parameters.object_pairs:
        data[fxobs.forecast] = EMPTY_DF['value']
        if fxobs.reference_forecast is not None:
            data[fxobs.reference_forecast] = _test_data[
                fxobs.reference_forecast.forecast_id]
        if isinstance(fxobs, datamodel.ForecastAggregate):
            data[fxobs.aggregate] = _test_data[fxobs.aggregate.aggregate_id]
        else:
            data[fxobs.observation] = _test_data[
                fxobs.observation.observation_id]

    raw = main.create_raw_report_from_data(report, data)
    assert isinstance(raw, datamodel.RawReport)
    assert len(raw.plots.figures) > 0


def test_create_raw_report_from_data_no_obs(mocker, report_objects,
                                            _test_data):
    report = report_objects[0]
    data = {}
    for fxobs in report.report_parameters.object_pairs:
        data[fxobs.forecast] = _test_data[fxobs.forecast.forecast_id]
        if fxobs.reference_forecast is not None:
            data[fxobs.reference_forecast] = _test_data[
                fxobs.reference_forecast.forecast_id]
        data[fxobs.data_object] = EMPTY_DF

    raw = main.create_raw_report_from_data(report, data)
    assert isinstance(raw, datamodel.RawReport)
    assert len(raw.plots.figures) > 0


def test_create_raw_report_from_data_no_data(mocker, report_objects):
    report = report_objects[0]
    data = {}
    for fxobs in report.report_parameters.object_pairs:
        data[fxobs.forecast] = EMPTY_DF['value']
        if fxobs.reference_forecast is not None:
            data[fxobs.reference_forecast] = EMPTY_DF['value']
        data[fxobs.data_object] = EMPTY_DF
    raw = main.create_raw_report_from_data(report, data)
    assert isinstance(raw, datamodel.RawReport)
    assert len(raw.plots.figures) > 0


def test_create_raw_report_from_data_event(mocker, event_report_objects,
                                           _test_event_data):
    report = event_report_objects[0]
    data = {}
    for fxobs in report.report_parameters.object_pairs:
        data[fxobs.forecast] = _test_event_data[fxobs.forecast.forecast_id]
        data[fxobs.observation] = _test_event_data[
            fxobs.observation.observation_id
        ]

    raw = main.create_raw_report_from_data(report, data)
    assert isinstance(raw, datamodel.RawReport)
    assert len(raw.plots.figures) > 0


def test_capture_report_failure(mocker):
    api_post = mocker.patch('solarforecastarbiter.io.api.APISession.post')

    def fail():
        raise TypeError()

    session = api.APISession('nope')
    failwrap = main.capture_report_failure('report_id', session)
    with pytest.raises(TypeError):
        failwrap(fail)()
    assert 'Critical ' in api_post.call_args_list[0][1]['json']['messages'][0]['message']  # NOQA
    assert api_post.call_args_list[1][0][0] == '/reports/report_id/status/failed'  # NOQA


def test_capture_report_failure_is_valid_datamodel(mocker):
    api_post = mocker.patch('solarforecastarbiter.io.api.APISession.post')

    def fail():
        raise TypeError()

    session = api.APISession('nope')
    failwrap = main.capture_report_failure('report_id', session)
    with pytest.raises(TypeError):
        failwrap(fail)()
    raw = datamodel.RawReport.from_dict(api_post.call_args_list[0][1]['json'])
    assert 'CRITICAL' == raw.messages[0].level


def test_capture_report_failure_msg(mocker):
    api_post = mocker.patch('solarforecastarbiter.io.api.APISession.post')

    def fail():
        raise TypeError()

    session = api.APISession('nope')
    failwrap = main.capture_report_failure('report_id', session)
    err_msg = 'Super bad error message'
    with pytest.raises(TypeError):
        failwrap(fail, err_msg=err_msg)()
    assert api_post.call_args_list[0][1]['json']['messages'][0]['message'] == err_msg  # NOQA
    assert api_post.call_args_list[1][0][0] == '/reports/report_id/status/failed'  # NOQA


def test_compute_report(mocker, report_objects, mock_data):
    report = report_objects[0]
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_report',
        return_value=report)
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_raw_report'
    )
    raw = main.compute_report('nope', report.report_id)
    assert isinstance(raw, datamodel.RawReport)
    assert len(raw.plots.figures) > 0


def test_compute_report_request_mock(
        mocker, report_objects, mock_data, requests_mock):
    report = report_objects[0]
    mocker.patch(
        'solarforecastarbiter.io.api.APISession.get_report',
        return_value=report)
    requests_mock.register_uri(
        'POST', re.compile('.*/reports/.*/values'),
        text=lambda *x: str(uuid.uuid1()))
    rep_post = requests_mock.register_uri(
        'POST', re.compile('.*/reports/.*/raw')
    )
    status_up = requests_mock.register_uri(
        'POST', re.compile('.*/reports/.*/status')
    )
    raw = main.compute_report('nope', report.report_id)
    assert isinstance(raw, datamodel.RawReport)
    assert len(raw.plots.figures) > 0
    assert rep_post.called
    assert 'complete' in status_up.last_request.path


@pytest.fixture()
def assert_post_called(mocker):
    post_report = mocker.patch(
        'solarforecastarbiter.io.api.APISession.post_raw_report')
    yield
    assert post_report.called
    assert post_report.call_args[0][0] == 'repid'
    assert post_report.call_args[0][-1] == 'failed'
    assert len(post_report.call_args[0][1].messages[0].message) > 0


def test_compute_report_get_report_fail(mocker, assert_post_called):
    mocker.patch('solarforecastarbiter.io.api.APISession.get_report',
                 side_effect=TypeError)
    with pytest.raises(TypeError):
        main.compute_report('nope', 'repid')


@pytest.fixture()
def get_report_mocked(mocker, report_objects):
    mocker.patch('solarforecastarbiter.io.api.APISession.get_report',
                 return_value=report_objects[0])


def test_compute_report_get_data_fail(
        mocker, get_report_mocked, assert_post_called):
    mocker.patch('solarforecastarbiter.reports.main.get_data_for_report',
                 side_effect=TypeError)
    with pytest.raises(TypeError):
        main.compute_report('nope', 'repid')


def test_compute_report_compute_fail(mocker, get_report_mocked, mock_data,
                                     assert_post_called):
    mocker.patch(
        'solarforecastarbiter.reports.main.create_raw_report_from_data',
        side_effect=TypeError)
    with pytest.raises(TypeError):
        main.compute_report('nope', 'repid')



import pandas as pd
from pandas.testing import assert_index_equal
import pytest


from matplotlib import cm
from matplotlib.colors import Normalize
from solarforecastarbiter.plotting import utils


@pytest.mark.parametrize('var,exp', [
    ('ghi', 'GHI (W/m^2)'),
    ('dc_power', 'DC Power (MW)')
])
def test_format_variable_name(var, exp):
    out = utils.format_variable_name(var)
    assert out == exp


@pytest.mark.parametrize('dobj,removal', [
    (pd.Series, slice(5, 10)),
    (pd.DataFrame, slice(12, 15))
])
def test_align_index(dobj, removal):
    index = pd.date_range(start='now', freq='5min',
                          periods=20, name='timestamp')
    data = dobj(index=index, dtype=float)
    data = data.drop(index[removal])
    out = utils.align_index(data, pd.Timedelta('5min'))
    assert_index_equal(out.index, index)


def test_align_index_new_length():
    index = pd.date_range(start='now', freq='5min',
                          periods=20, name='timestamp')
    data = pd.Series(index=index, dtype=float)
    out = utils.align_index(data, pd.Timedelta('1min'))
    nindex = pd.date_range(start=index[0], end=index[-1], freq='1min',
                           name='timestamp')
    assert_index_equal(out.index, nindex)


def test_align_index_limit():
    index = pd.date_range(start='now', freq='5min',
                          periods=20, name='timestamp')
    data = pd.Series(index=index, dtype=float)
    out = utils.align_index(data, pd.Timedelta('5min'),
                            limit=pd.Timedelta('60min'))
    nindex = pd.date_range(start=index[-13], end=index[-1], freq='5min',
                           name='timestamp')
    assert_index_equal(out.index, nindex)


@pytest.mark.parametrize('label,method', [
    ('instant', 'line'),
    ('beginning', 'step'),
    ('ending', 'step'),
    ('event', 'step'),
    pytest.param('other', '', marks=pytest.mark.xfail(raises=ValueError))
])
def test_line_or_step(label, method):
    out = utils.line_or_step(label)
    assert out[0] == method
    assert isinstance(out[1], dict)
    assert isinstance(out[2], dict)


@pytest.mark.parametrize('label', [
    'instant',
    'beginning',
    'ending',
    'event',
    pytest.param('other', marks=pytest.mark.xfail(raises=ValueError))
])
def test_line_or_step_plotly(label):
    out = utils.line_or_step_plotly(label)
    assert isinstance(out, dict)


color_map = cm.get_cmap('viridis')
color_scaler = cm.ScalarMappable(
    Normalize(vmin=0, vmax=1),
    color_map,
)


@pytest.mark.parametrize('percentile,expected', [
    (100, '#fde725'),
    (90, '#bddf26'),
    (50, '#21918c'),
    (20, '#414487'),
    (5, '#471365'),
    (1, '#450457'),
])
def test_distribution_fill_color(percentile, expected):
    assert utils.distribution_fill_color(color_scaler, percentile) == expected


@pytest.mark.parametrize('cvs,expected', [
    ([5, 20, 50, 80, 95], True),
    ([5.0, 20.0, 50.0, 80.0, 95.0], True),
    ([5.0, 50.0, 80.0, 95.0], False),
    ([5, 50, 8, 9], False),
    ([5, 10, 20, 30, 50, 60, 70, 75, 80, 90], False),
    ([5], False),
])
def test_percentiles_are_symmetric(cvs, expected):
    assert utils.percentiles_are_symmetric(cvs) == expected



import datetime as dt
import json


import bokeh
import numpy as np
import pandas as pd
import pytest


from solarforecastarbiter.plotting import timeseries


@pytest.mark.parametrize('times,exp', [
    ((pd.Timestamp('20190501T0000Z'), pd.Timestamp('20190501T1200Z')),
     'OBJECT 2019-05-01 00:00 to 2019-05-01 12:00 UTC'),
    ((dt.datetime(2019, 5, 1, 0, 0, tzinfo=dt.timezone.utc),
      dt.datetime(2019, 5, 1, 12, 0, tzinfo=dt.timezone.utc)),
     'OBJECT 2019-05-01 00:00 to 2019-05-01 12:00 UTC'),
    pytest.param((pd.Timestamp('20190501T0000'),
                  pd.Timestamp('20190501T1200Z')),
                 '', marks=pytest.mark.xfail(raises=ValueError)),
    pytest.param((pd.Timestamp('20190501T0000-0700'),
                  pd.Timestamp('20190501T1200Z')),
                 '', marks=pytest.mark.xfail(raises=ValueError)),
    pytest.param((dt.datetime.now(),
                  pd.Timestamp('20190501T1200Z')),
                 '', marks=pytest.mark.xfail(raises=ValueError))
])
def test_build_figure_title(times, exp):
    out = timeseries.build_figure_title(
        'OBJECT', *times)
    assert out == exp


TEST_FLAGS = pd.DataFrame({'USER FLAGGED': [False, False, True],
                           'NIGHTTIME': [True, False, False],
                           'CLEARSKY EXCEEDED': [None, None, None]},
                          index=pd.date_range(start='20190501T0000Z',
                                              freq='1h', periods=3))


def test_make_quality_bars():
    source = bokeh.models.ColumnDataSource(TEST_FLAGS)
    out = timeseries.make_quality_bars(
        source, 800, (TEST_FLAGS.index[0], TEST_FLAGS.index[-1]))
    assert isinstance(out, list)
    assert out[0].title.text == 'Quality Flags'
    assert len(out) == 2
    assert isinstance(out[0], bokeh.models.Plot)
    assert isinstance(out[1], bokeh.models.Plot)


@pytest.mark.parametrize('active', [True, False])
@pytest.mark.parametrize('addline', [True, False])
def test_add_hover_tool(addline, active):
    fig = bokeh.plotting.figure()
    if active:
        source = bokeh.models.ColumnDataSource({'active_flags': [[0]]})
    else:
        source = bokeh.models.ColumnDataSource()
    timeseries.add_hover_tool(fig, source, add_line=addline)
    assert len(fig.renderers) == int(addline)
    assert isinstance(fig.tools[-1], bokeh.models.HoverTool)
    if active:
        assert fig.tools[-1].tooltips[-1][0] == 'quality flags'
    else:
        assert len(fig.tools[-1].tooltips) == 2


@pytest.mark.parametrize('df', [
    pd.DataFrame(
        index=pd.date_range(start='now', freq='1min',
                            periods=2, tz='UTC', name='timestamp')),
    pytest.param(pd.DataFrame(), marks=pytest.mark.xfail(raises=KeyError)),
    pytest.param(pd.DataFrame(index=pd.date_range(
        start='now', freq='1min', periods=0, name='timestamp')),
                 marks=pytest.mark.xfail(raises=IndexError))
])
@pytest.mark.parametrize('label', ['instant', 'beginning', 'ending'])
def test_make_basic_timeseries(label, df):
    source = bokeh.models.ColumnDataSource(df)
    fig = timeseries.make_basic_timeseries(source, 'OBJECT', 'ghi',
                                           label, 800)

    if label == 'instant':
        assert len(fig.renderers) == 1
    else:
        assert len(fig.renderers) == 2  # has hover line

    assert 'Time' in fig.xaxis[0].axis_label
    assert 'GHI' in fig.yaxis[0].axis_label
    assert 'OBJECT'in fig.title.text


def test_make_basic_event_timeseries():
    df = pd.DataFrame(
        index=pd.date_range(start='now', freq='1min',
                            periods=2, tz='UTC', name='timestamp'))
    source = bokeh.models.ColumnDataSource(df)
    fig = timeseries.make_basic_timeseries(source, 'OBJECT', 'event',
                                           'event', 800)

    assert fig.yaxis.ticker.ticks == [0, 1]
    assert list(fig.yaxis.major_label_overrides.values()) == ['True', 'False']
    assert 'Time' in fig.xaxis[0].axis_label
    assert 'Event' in fig.yaxis[0].axis_label
    assert 'OBJECT'in fig.title.text


def test_generate_forecast_figure(ac_power_forecast_metadata,
                                  forecast_values):
    # not great, just testing that script and div are returned
    out = timeseries.generate_forecast_figure(ac_power_forecast_metadata,
                                              forecast_values)
    assert isinstance(out, bokeh.models.Column)


def test_generate_forecast_figure_components(ac_power_forecast_metadata,
                                             forecast_values):
    # not great, just testing that script and div are returned
    out = timeseries.generate_forecast_figure(ac_power_forecast_metadata,
                                              forecast_values,
                                              return_components=True)
    assert '<script' in out[0]
    assert '<div' in out[1]


@pytest.mark.parametrize('rc', [True, False])
def test_generate_forecast_figure_empty(ac_power_forecast_metadata, rc):
    assert timeseries.generate_forecast_figure(ac_power_forecast_metadata,
                                               pd.Series(dtype=float),
                                               return_components=rc) is None


def test_generate_observation_figure(validated_observation_values,
                                     ac_power_observation_metadata_label):
    out = timeseries.generate_observation_figure(
        ac_power_observation_metadata_label, validated_observation_values)
    assert isinstance(out, bokeh.models.Column)


def test_generate_observation_figure_components(
        validated_observation_values, ac_power_observation_metadata_label):
    out = timeseries.generate_observation_figure(
        ac_power_observation_metadata_label, validated_observation_values,
        return_components=True)
    assert '<script' in out[0]
    assert '<div' in out[1]


def test_generate_observation_figure_missing_data(
        validated_observation_values, ac_power_observation_metadata):
    validated_observation_values.iloc[1] = [None, None]
    out = timeseries.generate_observation_figure(
        ac_power_observation_metadata, validated_observation_values,
        return_components=True)
    assert '<script' in out[0]
    assert 'MISSING' in out[0]
    assert '<div' in out[1]


@pytest.mark.parametrize('rc', [True, False])
def test_generate_observation_figure_empty(ghi_observation_metadata, rc):
    assert timeseries.generate_observation_figure(ghi_observation_metadata,
                                                  pd.DataFrame(),
                                                  return_components=rc) is None


@pytest.fixture
def prob_forecast_random_data():
    def f(forecast):
        frequency = pd.tseries.frequencies.to_offset(forecast.interval_length)
        start = pd.Timestamp('2020-01-01T00:00Z')
        end = pd.Timestamp('2020-01-03T00:00Z')
        idx = pd.date_range(start, end, freq=frequency)
        df = pd.DataFrame(index=idx)
        for cv in [c.constant_value for c in forecast.constant_values]:
            df[str(cv)] = np.random.rand(idx.size)
        return df
    return f


def test_generate_probabilistic_forecast_figure_x_forecast(
        prob_forecasts, prob_forecast_random_data):
    values = prob_forecast_random_data(prob_forecasts)
    fig = timeseries.generate_probabilistic_forecast_figure(
        prob_forecasts, values)
    assert fig['layout']['title']['text'] == 'DA GHI 2020-01-01 00:00 to 2020-01-03 00:00 UTC'  # NOQA: E501
    assert fig['layout']['xaxis']['title']['text'] == 'Time (UTC)'
    assert fig['layout']['yaxis']['title']['text'] == 'Probability (%)'
    fig_data = fig['data']
    assert len(fig_data) == 1
    assert len(fig_data[0]['x']) == values.index.size
    assert len(fig_data[0]['y']) == values.index.size
    assert fig_data[0]['showlegend']


def test_generate_probabilistic_forecast_figure_y_forecast(
        prob_forecasts_y,
        prob_forecast_constant_value_y_factory,
        prob_forecast_random_data,
        ):
    new_constant_values = [prob_forecast_constant_value_y_factory(5.0)]
    prob_forecast = prob_forecasts_y.replace(
        constant_values=new_constant_values)
    values = prob_forecast_random_data(prob_forecast)
    fig = timeseries.generate_probabilistic_forecast_figure(
        prob_forecasts_y, values)
    assert fig['layout']['title']['text'] == 'DA GHI 2020-01-01 00:00 to 2020-01-03 00:00 UTC'  # NOQA: E501
    assert fig['layout']['xaxis']['title']['text'] == 'Time (UTC)'
    assert fig['layout']['yaxis']['title']['text'] == 'GHI (W/m^2)'
    fig_data = fig['data']
    assert len(fig_data) == 1
    assert len(fig_data[0]['x']) == values.index.size
    assert len(fig_data[0]['y']) == values.index.size
    assert not fig_data[0]['showlegend']


@pytest.fixture
def prob_forecast_constant_value_y_factory(
        prob_forecast_constant_value_y_text,
        _prob_forecast_constant_value_from_dict):
    def f(new_constant_value):
        fx_dict = json.loads(prob_forecast_constant_value_y_text)
        fx_dict['constant_value'] = new_constant_value
        return _prob_forecast_constant_value_from_dict(fx_dict)
    return f


def test_generate_probabilistic_forecast_figure_y_forecast_symmetric(
        prob_forecasts_y,
        prob_forecast_constant_value_y_factory,
        prob_forecast_random_data,
        ):
    new_constant_values = [prob_forecast_constant_value_y_factory(x)
                           for x in [5.0, 10.0, 50.0, 90.0, 95.0]]
    prob_forecast = prob_forecasts_y.replace(
        constant_values=new_constant_values)
    values = prob_forecast_random_data(prob_forecast)
    fig = timeseries.generate_probabilistic_forecast_figure(
        prob_forecasts_y, values)
    assert fig['layout']['title']['text'] == 'DA GHI 2020-01-01 00:00 to 2020-01-03 00:00 UTC'  # NOQA: E501
    assert fig['layout']['xaxis']['title']['text'] == 'Time (UTC)'
    assert fig['layout']['yaxis']['title']['text'] == 'GHI (W/m^2)'
    fig_data = fig['data']
    assert len(fig_data) == 5
    for trace in fig_data:
        assert len(trace['x']) == values.index.size
        assert len(trace['y']) == values.index.size
    assert fig_data[0]['fill'] is None
    for trace in fig_data[1:]:
        assert trace['fill'] == 'tonexty'


def test_generate_probabilistic_forecast_figure_empty_values(
        prob_forecasts_y, prob_forecast_random_data):
    values = pd.DataFrame()
    fig = timeseries.generate_probabilistic_forecast_figure(
        prob_forecasts_y, values)
    assert fig is None



from functools import partial
import json
import logging
import os
from pkg_resources import resource_filename, Requirement


import pandas as pd
from pytz.exceptions import NonExistentTimeError


from solarforecastarbiter.datamodel import Observation
from solarforecastarbiter.io.fetch import pvdaq
from solarforecastarbiter.io.reference_observations import (
    common, default_forecasts)


logger = logging.getLogger('reference_data')


DEFAULT_SITEFILE = resource_filename(
    Requirement.parse('solarforecastarbiter'),
    'solarforecastarbiter/io/reference_observations/'
    'pvdaq_reference_sites.json')


def adjust_site_parameters(site):
    """Kludge the extra metadata in a json file into the metadata dict
    derived from a csv file.

    Parameters
    ----------
    site: dict

    Returns
    -------
    dict
        Copy of input plus a new key 'modeling_parameters' and more
        metadata in extra_parameters.

    See also
    --------
    solarforecastarbiter.io.reference_observations.site_df_to_dicts
    """
    return common.apply_json_site_parameters(DEFAULT_SITEFILE, site)


def initialize_site_observations(api, site):
    """Creates an observation at the site for each variable in the PVDAQ
    site's file.

    Parameters
    ----------
    api : io.api.APISession
        API Session object, authenticated for the Reference user.
    site : datamodel.Site
        The site object for which to create the Observations.
    """
    try:
        extra_params = common.decode_extra_parameters(site)
    except ValueError:
        logger.warning('Cannot create reference observations at PVDAQ site '
                       f'{site.name}, missing required parameters.')
        return
    site_api_id = int(extra_params['network_api_id'])
    with open(DEFAULT_SITEFILE) as fp:
        obs_metadata = json.load(fp)['observations']

    for obs in obs_metadata:
        obs_extra_params = json.loads(obs['extra_parameters'])
        if obs_extra_params['network_api_id'] == site_api_id:
            obs['site'] = site
            observation = Observation.from_dict(obs)
            common.check_and_post_observation(api, observation)


def initialize_site_forecasts(api, site):
    """
    Create a forecasts for each variable measured at the site

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An active Reference user session.
    site : datamodel.Site
        The site object for which to create Forecasts.
    """
    try:
        extra_params = common.decode_extra_parameters(site)
    except ValueError:
        logger.warning('Cannot create reference observations at PVDAQ site '
                       f'{site.name}, missing required parameters.')
        return
    site_api_id = int(extra_params['network_api_id'])
    with open(DEFAULT_SITEFILE) as fp:
        obs_metadata = json.load(fp)['observations']

    obs_vars = []
    for obs in obs_metadata:
        obs_extra_params = json.loads(obs['extra_parameters'])
        if obs_extra_params['network_api_id'] == site_api_id:
            obs_vars.append(obs['variable'])

    common.create_forecasts(
        api, site, obs_vars, default_forecasts.TEMPLATE_FORECASTS)


def fetch(api, site, start, end, *, nrel_pvdaq_api_key):
    """Retrieve observation data for a PVDAQ site between start and end.

    Parameters
    ----------
    api : io.APISession
        Unused but conforms to common.update_site_observations call
    site : datamodel.Site
        Site object with the appropriate metadata.
    start : datetime
        The beginning of the period to request data for.
    end : datetime
        The end of the period to request data for.
    nrel_pvdaq_api_key : str
        API key for developer.nrel.gov

    Returns
    -------
    data : pandas.DataFrame
        All of the requested data concatenated into a single DataFrame.
    """
    try:
        site_extra_params = common.decode_extra_parameters(site)
    except ValueError:
        return pd.DataFrame()
    try:
        years = list(range(start.year, end.year + 1))
        obs_df = pvdaq.get_pvdaq_data(
            site_extra_params['network_api_id'], years,
            api_key=nrel_pvdaq_api_key)
    except Exception:
        # Not yet sure what kind of errors we might hit in production
        logger.warning(f'Could not retrieve data for site {site.name}'
                       f' between {start} and {end}.')
        return pd.DataFrame()
    obs_df = _watts_to_mw(obs_df)
    try:
        obs_df = obs_df.tz_localize(site.timezone)
    except NonExistentTimeError as e:
        logger.warning(f'Could not localize data for site {site.name} '
                       f'due to DST issue: {e}')
        return pd.DataFrame()
    return obs_df


def _watts_to_mw(obs_df):
    obs_df_power_mw = obs_df.filter(regex='.*[pP]ower.*') / 1e6
    obs_df[obs_df_power_mw.columns] = obs_df_power_mw
    return obs_df


def update_observation_data(api, sites, observations, start, end):
    """Post new observation data to all PVDAQ observations from
    start to end.

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An active Reference user session.
    sites: list
        List of all reference sites as Objects
    observations: list of solarforecastarbiter.datamodel.Observation
        List of all reference observations.
    start : datetime
        The beginning of the period to request data for.
    end : datetime
        The end of the period to request data for.

    Raises
    ------
    KeyError
        If NREL_PVDAQ_API_KEY environmental variable is not set.
        Abuse of KeyError - should probably be ValueError - but kept for
        consistency with other reference_observations modules.
    """
    nrel_pvdaq_api_key = os.getenv('NREL_PVDAQ_API_KEY')
    if nrel_pvdaq_api_key is None:
        raise KeyError('"NREL_PVDAQ_API_KEY" environment variable must be '
                       'set to update PVDAQ observation data.')
    pvdaq_sites = common.filter_by_networks(sites, ['NREL PVDAQ'])
    for site in pvdaq_sites:
        common.update_site_observations(
            api, partial(fetch, nrel_pvdaq_api_key=nrel_pvdaq_api_key),
            site, observations, start, end)



"""Contains a config dictionary mapping SFA variable names to MIDC fields.
The values in these fields are as they would appear returned from the
`pvlib.iotools.read_midc_raw_data_from_nrel` function.

All site's field list can be found at:
https://midcdmz.nrel.gov/apps/daily.pl?site=<SITE ID>&live=1
Where id is the key found in this dictionary
"""

midc_var_map = {
    'BMS': {
        'ghi': 'Global CMP22 (vent/cor) [W/m^2]',
        'dni': 'Direct NIP [W/m^2]',
        'dhi': 'Diffuse CM22-1 (vent/cor) [W/m^2]',
        'wind_speed': 'Avg Wind Speed @ 33ft [m/s]',
        'air_temperature': 'Tower Dry Bulb Temp [deg C]',
        'relative_humidity': 'Tower RH [%]',
    },
    'UOSMRL': {
        'ghi': 'Global CMP22 [W/m^2]',
        'dni': 'Direct NIP [W/m^2]',
        'dhi': 'Diffuse Schenk [W/m^2]',
        'air_temperature': 'Air Temperature [deg C]',
        'relative_humidity': 'Relative Humidity [%]',
        'wind_speed': 'Avg Wind Speed @ 10m [m/s]',
    },
    'HSU': {
        'ghi': 'Global Horiz [W/m^2]',
        'dni': 'Direct Normal (calc) [W/m^2]',
        'dhi': 'Diffuse Horiz (band corr) [W/m^2]',
    },
    'UTPASRL': {
        'ghi': 'Global Horizontal [W/m^2]',
        'dni': 'Direct Normal [W/m^2]',
        'dhi': 'Diffuse Horizontal [W/m^2]',
        'air_temperature': 'CHP1 Temp [deg C]',
    },
    'UAT': {
        'ghi': 'Global Horiz (platform) [W/m^2]',
        'dni': 'Direct Normal [W/m^2]',
        'dhi': 'Diffuse Horiz [W/m^2]',
        'air_temperature': 'Air Temperature [deg C]',
        'relative_humidity': 'Rel Humidity [%]',
        'wind_speed': 'Avg Wind Speed @ 3m [m/s]',
    },
    'STAC': {
        'ghi': 'Global Horizontal [W/m^2]',
        'dni': 'Direct Normal [W/m^2]',
        'dhi': 'Diffuse Horizontal [W/m^2]',
        'wind_speed': 'Avg Wind Speed @ 10m [m/s]',
        'air_temperature': 'Air Temperature [deg C]',
        'relative_humidity': 'Rel Humidity [%]',
    },
    'UNLV': {
        'ghi': 'Global Horiz [W/m^2]',
        'dni': 'Direct Normal [W/m^2]',
        'dhi': 'Diffuse Horiz (calc) [W/m^2]',
        'air_temperature': 'Dry Bulb Temp [deg C]',
        'wind_speed': 'Avg Wind Speed @ 30ft [m/s]',
    },
    'ORNL': {
        'ghi': 'Global Horizontal [W/m^2]',
        'dni': 'Direct Normal [W/m^2]',
        'dhi': 'Diffuse Horizontal [W/m^2]',
        'air_temperature': 'Air Temperature [deg C]',
        'relative_humidity': 'Rel Humidity [%]',
        'wind_speed': 'Avg Wind Speed @ 42ft [m/s]',
    },
    'NELHA': {
        'ghi': 'Global Horizontal [W/m^2]',
        'air_temperature': 'Air Temperature [deg C]',
        'wind_speed': 'Avg Wind Speed @ 10m [m/s]',
        'relative_humidity': 'Rel Humidity [%]',
    },
    'ULL': {
        'ghi': 'Global Horizontal [W/m^2]',
        'dni': 'Direct Normal [W/m^2]',
        'dhi': 'Diffuse Horizontal [W/m^2]',
        'air_temperature': 'Air Temperature [deg C]',
        'relative_humidity': 'Rel Humidity [%]',
        'wind_speed': 'Avg Wind Speed @ 3m [m/s]',
    },
    'VTIF': {
        'ghi': 'Global Horizontal [W/m^2]',
        'dni': 'Direct Normal [W/m^2]',
        'dhi': 'Diffuse Horizontal [W/m^2]',
        'air_temperature': 'Air Temperature [deg C]',
        'wind_speed': 'Avg Wind Speed @ 3m [m/s]',
        'relative_humidity': 'Rel Humidity [%]'
    },
    'NWTC': {
        'ghi': 'Global PSP [W/m^2]',
        'air_temperature': 'Temperature @ 2m [deg C]',
        'wind_speed': 'Avg Wind Speed @ 10m [m/s]',
        'relative_humidity': 'Relative Humidity [%]',
    },
}



"""Collection of code for requesting and parsing ARM data.
Documentation for the ARM Live Data Web Service can be found
here: https://adc.arm.gov/armlive/
"""
import json
import logging
import netCDF4
import pandas as pd
import requests
import time


logger = logging.getLogger(__name__)

ARM_FILES_LIST_URL = 'https://adc.arm.gov/armlive/data/query'
ARM_FILES_DOWNLOAD_URL = 'https://adc.arm.gov/armlive/data/saveData'


# These lists are the commonly available irradiance and meteorological
# variables found in ARM data. Users can import and pass these to fetch_arm
# to parse out these variables.
IRRAD_VARIABLES = ['down_short_hemisp', 'down_short_diffuse_hemisp',
                   'short_direct_normal']
MET_VARIABLES = ['temp_mean', 'rh_mean', 'wspd_arith_mean']


def format_date(date_object):
    return date_object.strftime('%Y-%m-%d')


def request_arm_file_list(user_id, api_key, datastream, start, end):
    """Make an http request to the ARM live API for filenames between start
    and end.

    Parameters
    ----------
    user_id: string
        ARM user id.
    api_key: string
        ARM live API access token.
    datastream: string
        Name of the datastream to query for files.
    start: datetime
        Beginning of period for which to request data.
    end: datetime
        End of period for which to request data.

    Returns
    -------
    dict
        The json response parsed into a dictionary.
    """
    params = {'user': f'{user_id}:{api_key}',
              'ds': datastream,
              'start': format_date(start),
              'end': format_date(end),
              'wt': 'json'}
    response = requests.get(ARM_FILES_LIST_URL, params=params)
    return json.loads(response.text)


def list_arm_filenames(user_id, api_key, datastream, start, end):
    """Get a list of filenames from ARM for the given datastream between
    start and end.

    Parameters
    ----------
    user_id: string
        ARM user id.
    api_key: string
        ARM live API access token.
    datastream: string
        Name of the datastream to query for files.
    start: datetime
        Beginning of period for which to request data.
    end: datetime
        End of period for which to request data.

    Returns
    -------
    list
        List of filenames as strings.
    """
    response = request_arm_file_list(user_id, api_key,
                                     datastream, start, end)
    return response['files']


def request_arm_file(user_id, api_key, filename, retries=5):
    """Get a file from ARM live in the form of a stream so that the python netCDF4
    module can read it.

    Parameters
    ----------
    user_id: string
        ARM user id.
    api_key: string
        ARM live API access token.
    filename: string
        Filename to request
    retries: int
        Number of attempts remaining to successfully request data after
        ChunkedEncodingError.

    Returns
    -------
    stream
        The API response in the form of a stream to be consumed by
        netCDF4.Dataset().

    Raises
    ------
    request.exceptions.ChunkedEncodingError
        Reraises this error when all retries are exhausted.
    """
    params = {'user': f'{user_id}:{api_key}',
              'file': filename}

    try:
        request = requests.get(ARM_FILES_DOWNLOAD_URL, params=params)
    except requests.exceptions.ChunkedEncodingError:
        if retries > 0:
            logger.debug(f'Retrying DOE ARM file {filename}: {retries}'
                         'remaining.')
            time.sleep((5 - retries) * 0.1)
            return request_arm_file(user_id, api_key, filename, retries-1)
        else:
            logger.warning(f'Requesting ARM file {filename} failed')
            raise
    nc_data = request.content
    return nc_data


def retrieve_arm_dataset(user_id, api_key, filename):
    """Request a file from the ARM Live API and return a netCDF4 Dataset.

    Parameters
    ----------
    user_id: string
        ARM user id.
    api_key: string
        ARM live API access token.
    filename: string
        Filename to request

    Returns
    -------
    netCDF4.Dataset
        Dataset of the API response.
    """
    nc_data = request_arm_file(user_id, api_key, filename)
    nc_file = netCDF4.Dataset(f'/tmp/{filename}', mode='r',
                              memory=nc_data)
    return nc_file


def extract_arm_variables(nc_file, variables):
    """Extracts variables and datetime index from an ARM netcdf.

    Parameters
    ----------
    nc_file: netCDF4 Dataset
        The ARM file read into a Dataset.
    variables: list
        List of string variable names to parse from the files.

    Returns
    -------
    DataFrame
        A pandas DataFrame with a column for each requested variable
        found in the ARM netcdf file, indexed by timestamp in UTC. If
        none of the requested variables are found, an empty DataFrame
        is returned.
    """
    var_data = {}
    for var in variables:
        try:
            var_data[var] = nc_file[var][:]
        except IndexError:
            continue
    if var_data:
        base_time = nc_file['base_time'][0].data.item()
        delta_time = nc_file['time'][:]
        times = pd.to_datetime(base_time + delta_time, unit='s', utc=True)
        return pd.DataFrame(index=times, data=var_data)
    else:
        return pd.DataFrame()


def fetch_arm(user_id, api_key, datastream, variables, start, end):
    """Gets data from ARM API and concatenates requested datastreams into
    a single Pandas Dataframe.

    Parameters
    ----------
    user_id: string
        ARM user id.
    api_key: string
        ARM live API access token.
    datastream: string
        The datastream to request.
    variables
        List of variables to parse from the datastream.
    start: datetime
        The start of the interval to request data for.
    end: datetime
        The end of the interval to request date for.

    Returns
    -------
    DataFrame
        A DataFrame containing all of the available variables over the
        requested period.

    Notes
    -----
    Elements of the variable list that are not found in the datastream
    are ignored, this is to allow iteration over many datastreams without
    knowing their exact contents. If none of the requested variables are
    found, an empty DataFrame will be returned. Users should verify the
    contents of the return value before use.

    Example
    -------
    A user requesting data for the variables 'down_short_hemisp' and
    'short_direct_normal' from the datastream 'sgpqcrad1longC1.c1' for
    the days between 2019-02-27 and 2019-03-01 could expect the
    following DataFrame.

    .. code::

                                  down_short_hemisp  short_direct_normal
       2019-02-27 00:00:00+00:00           7.182889            -1.399250
       2019-02-27 00:01:00+00:00           6.943601            -1.317890
       2019-02-27 00:02:00+00:00           6.686488            -1.235140
       ...
       2019-03-01 23:57:00+00:00           6.943601            -1.317890
       2019-03-01 23:58:00+00:00           6.686488            -1.235140
       2019-03-01 23:59:00+00:00           6.395981            -1.226730
    """
    datastream_dfs = []
    filenames = list_arm_filenames(user_id, api_key, datastream, start, end)
    for filename in filenames:
        try:
            nc_file = retrieve_arm_dataset(user_id, api_key, filename)
        except requests.exceptions.ChunkedEncodingError:
            logger.error(f'Request failed for DOE ARM file {filename}')
        else:
            datastream_df = extract_arm_variables(nc_file, variables)
            datastream_dfs.append(datastream_df)
    if len(datastream_dfs) > 0:
        new_data = pd.concat(datastream_dfs)
        return new_data
    else:
        return pd.DataFrame()
